from crystal.tests.test_shell import (
    # TODO: Extract methods to shared module rather than importing private functions
    _create_new_empty_project,
    _OK_THREAD_STOP_SUFFIX,
    _py_eval,
    crystal_shell,
)
from crystal.tests.util.windows import OpenOrCreateDialog
from crystal.tests.util.xos import skip_on_windows
import tempfile
import textwrap


async def test_can_close_project_with_menuitem() -> None:
    async with (await OpenOrCreateDialog.wait_for()).create(autoclose=False) as (mw, _):
        await mw.close_with_menuitem()


async def test_can_quit_with_menuitem() -> None:
    skip_on_windows()()  # crystal_shell doesn't support Windows
    
    with crystal_shell() as (crystal, _):
        _create_new_empty_project(crystal)
        
        _py_eval(crystal, textwrap.dedent(f'''\
            from crystal.tests.util.runner import run_test
            from crystal.tests.util.windows import MainWindow
            from threading import Thread
            import wx

            async def quit_with_menuitem():
                mw = await MainWindow.wait_for()
                #
                await mw.quit_with_menuitem()
                #
                print('OK')

            t = Thread(target=lambda: run_test(quit_with_menuitem))
            t.start()
            '''), stop_suffix=('crystal.util.xthreading.NoForegroundThreadError\n',))


async def test_can_open_preferences_with_menuitem() -> None:
    async with (await OpenOrCreateDialog.wait_for()).create() as (mw, _):
        prefs_dialog = await mw.open_preferences_with_menuitem()
        await prefs_dialog.ok()



from contextlib import (
    AbstractContextManager, asynccontextmanager, contextmanager, nullcontext,
)
from crystal.browser.tasktree import _MoreNodeView, TaskTreeNode
from crystal.model import Resource
import crystal.task
from crystal.task import (
    DownloadResourceTask,
    DownloadResourceGroupMembersTask, DownloadResourceGroupTask, 
    _PlaceholderTask, SCHEDULING_STYLE_ROUND_ROBIN,
    Task, UpdateResourceGroupMembersTask,
)
from crystal.tests.util.asserts import assertEqual
from crystal.tests.util.controls import click_button, TreeItem
from crystal.tests.util.data import MAX_TIME_TO_DOWNLOAD_404_URL
from crystal.tests.util.downloads import load_children_of_drg_task
from crystal.tests.util.server import served_project
from crystal.tests.util.tasks import (
    append_deferred_top_level_tasks,
    mark_as_complete as _mark_as_complete,
    scheduler_disabled, scheduler_thread_context,
    ttn_for_task as _ttn_for_task,
)
from crystal.tests.util.wait import tree_has_no_children_condition, wait_for, wait_while
from crystal.tests.util.windows import MainWindow, OpenOrCreateDialog
from crystal.ui.tree import NodeView
from crystal.ui.tree2 import NodeView as NodeView2
from crystal.util.xcollections.lazy import (
    AppendableLazySequence, UnmaterializedItem,
)
from crystal.model import Project, Resource, ResourceGroup, RootResource
import math
import tempfile
from typing import AsyncIterator, Callable, cast, Iterator, List, Optional, Sequence, Tuple
from unittest import skip
from unittest.mock import call, patch


# === Test: SCHEDULING_STYLE_SEQUENTIAL tasks: Limit visible children ===

async def test_when_start_downloading_large_group_then_show_100_children_plus_trailing_more_node() -> None:
    N = 3
    M = 2
    async with _project_with_resource_group_starting_to_download(
                resource_count=N + M + 1,
                small_max_visible_children=N,
                small_max_leading_complete_children=M,
                scheduler_thread_enabled=False,
            ) as (mw, download_rg_ttn, download_rg_members_ttn, _):
        project = Project._last_opened_project
        assert project is not None
        
        parent_ttn = download_rg_members_ttn
        
        # Ensure only first 100 member download tasks are visible,
        # with a "# more" placeholder at the end
        assert 0 == _viewport_offset(parent_ttn)
        assert 0 + N == _materialized_child_task_count(parent_ttn)
        assert 0 == _unmaterialized_child_task_count(parent_ttn)
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == N + 1
        assert all([_is_download_resource_node(tn) for tn in children_tns[:-1]])
        assert _is_more_node(children_tns[-1])
        assert M + 1 == _value_of_more_node(children_tns[-1])
        
        # Ensure leading "# more" placeholder inserted after M+1 tasks complete
        if True:
            for i in range(M):
                _mark_as_complete(children_tasks[i])
                assert 0 == _viewport_offset(parent_ttn)
                assert 0 + N == _materialized_child_task_count(parent_ttn)
                assert 0 == _unmaterialized_child_task_count(parent_ttn)
                (children_tns, children_tasks) = \
                    (parent_ttn.tree_node.children, parent_ttn.task.children)
                assert len(children_tns) == N + 1
                assert all([_is_download_resource_node(tn) for tn in children_tns[:-1]])
                assert _is_more_node(children_tns[-1])
                assert M + 1 == _value_of_more_node(children_tns[-1])
            
            _mark_as_complete(children_tasks[M])
            assert 1 == _viewport_offset(parent_ttn)
            assert 1 + N == _materialized_child_task_count(parent_ttn)
            assert 1 == _unmaterialized_child_task_count(parent_ttn)
            (children_tns, children_tasks) = \
                (parent_ttn.tree_node.children, parent_ttn.task.children)
            assert len(children_tns) == 1 + N + 1
            assert _is_more_node(children_tns[0])
            assert all([_is_download_resource_node(tn) for tn in children_tns[1:-1]])
            assert _is_more_node(children_tns[-1])
            assert 1 == _value_of_more_node(children_tns[0])
            assert M - 1 >= 1
            assert M + 1 - 1 == _value_of_more_node(children_tns[-1])
        
        # Ensure leading/trailing "# more" placeholder increments/decrements
        # as each task completes, until trailing "# more" placeholder disappears
        for i in range(M+1, M+M):
            _mark_as_complete(children_tasks[i])
            assert (i - M + 1) == _viewport_offset(parent_ttn)
            assert (i - M + 1) + N == _materialized_child_task_count(parent_ttn)
            assert (i - M + 1) == _unmaterialized_child_task_count(parent_ttn)
            (children_tns, children_tasks) = \
                (parent_ttn.tree_node.children, parent_ttn.task.children)
            if M + 1 - (i - M + 1) >= 1:
                assert len(children_tns) == 1 + N + 1
                assert _is_more_node(children_tns[0])
                assert all([_is_download_resource_node(tn) for tn in children_tns[1:-1]])
                assert _is_more_node(children_tns[-1])
                assert (i - M + 1) == _value_of_more_node(children_tns[0])
                assert M + 1 - (i - M + 1) == _value_of_more_node(children_tns[-1])
            else:
                assert len(children_tns) == 1 + N
                assert _is_more_node(children_tns[0])
                assert all([_is_download_resource_node(tn) for tn in children_tns[1:]])
                assert (i - M + 1) == _value_of_more_node(children_tns[0])
                assert M + 1 - (i - M + 1) == 0
        
        # Ensure leading "# more" placeholder increments as each task completes,
        # until all tasks complete
        if True:
            # Complete remaining tasks up to but not including the last task
            for (i, j) in zip(range(M+M, M+N), range(0, M)):
                _mark_as_complete(parent_ttn.task.children[i])
                assert (i - M + 1) == _viewport_offset(parent_ttn)
                assert M + N + 1 == _materialized_child_task_count(parent_ttn)
                assert (i - M + 1) == _unmaterialized_child_task_count(parent_ttn)
                (children_tns, children_tasks) = \
                    (parent_ttn.tree_node.children, parent_ttn.task.children)
                assert len(children_tns) == 1 + N - j
                assert _is_more_node(children_tns[0])
                assert all([_is_download_resource_node(tn) for tn in children_tns[1:]])
                assert (i - M + 1) == _value_of_more_node(children_tns[0])
            
            # Complete the last task
            assert parent_ttn.task.children[-2].complete
            assert not parent_ttn.task.children[-1].complete
            _mark_as_complete(parent_ttn.task.children[-1])
            assert parent_ttn.task.complete


@skip('covered by: test_when_start_downloading_large_group_then_show_100_children_plus_trailing_more_node')
async def test_given_downloading_large_group_and_many_uncompleted_children_remaining_when_child_completes_then_show_leading_more_node_plus_100_children_plus_trailing_more_node() -> None:
    pass


@skip('covered by: test_when_start_downloading_large_group_then_show_100_children_plus_trailing_more_node')
async def test_given_downloading_large_group_and_few_uncompleted_children_remaining_when_child_completes_then_show_leading_more_node_plus_up_to_100_children_but_no_trailing_more_node() -> None:
    pass


@skip('covered by: test_when_start_downloading_large_group_then_show_100_children_plus_trailing_more_node')
async def test_while_downloading_large_group_then_keeps_no_more_than_100_member_download_tasks_in_memory() -> None:
    # The referenced covering test ensures that:
    #     _materialized_child_task_count(...) - _unmaterialized_child_task_count(...) <= N
    # for any N, including 100
    pass


async def test_given_group_has_leading_completed_children_when_start_downloading_large_group_then_shows_no_more_then_5_leading_completed_children() -> None:
    N = 5
    M = 3
    
    with _children_marked_as_complete_upon_creation(list(range(1, (M + 1) + 1))):
        async with _project_with_resource_group_starting_to_download(
                    resource_count=N + M + 1,
                    small_max_visible_children=N,
                    small_max_leading_complete_children=M,
                    scheduler_thread_enabled=False,
                ) as (mw, download_rg_ttn, download_rg_members_ttn, _):
            parent_ttn = download_rg_members_ttn
            
            assertEqual(1, _viewport_offset(parent_ttn))
            assertEqual(N, _viewport_length(parent_ttn))


async def test_given_group_has_more_leading_completed_children_than_visible_children_when_start_downloading_then_moves_viewport_to_appropriate_location() -> None:
    N = 5
    M = 3
    
    with _children_marked_as_complete_upon_creation(list(range(1, (N + 1) + 1))):
        async with _project_with_resource_group_starting_to_download(
                    resource_count=N + M + 1,
                    small_max_visible_children=N,
                    small_max_leading_complete_children=M,
                    scheduler_thread_enabled=False,
                ) as (mw, download_rg_ttn, download_rg_members_ttn, _):
            parent_ttn = download_rg_members_ttn
            
            assertEqual(3, _viewport_offset(parent_ttn))
            assertEqual(N, _viewport_length(parent_ttn))
            
            # Ensure does not crash when update Task._next_child_index
            # in way that passes through children that were unmaterialized
            # (and which should be assumed to be complete)
            # 
            # In particular ensure UnmaterializedItemError is handled correctly.
            parent_ttn.task.try_get_next_task_unit()
    
    with _children_marked_as_complete_upon_creation(list(range(1, (N + M + 1) + 1))):
        async with _project_with_resource_group_starting_to_download(
                    resource_count=N + M + 2,
                    small_max_visible_children=N,
                    small_max_leading_complete_children=M,
                    scheduler_thread_enabled=False,
                ) as (mw, download_rg_ttn, download_rg_members_ttn, _):
            parent_ttn = download_rg_members_ttn
            
            assertEqual(6, _viewport_offset(parent_ttn))
            assertEqual(M + 1, _viewport_length(parent_ttn))
            
            # Ensure does not crash when update
            # (first_more_node, intermediate_nodes, last_more_node)
            # in way that passes through children that were unmaterialized
            # (and which should be assumed to be complete)
            # 
            # In particular ensure UnmaterializedItem is handled correctly.
            parent_ttn.task.try_get_next_task_unit()


# === Test: SCHEDULING_STYLE_SEQUENTIAL tasks: Limit leading completed children ===

async def test_given_showing_less_than_5_leading_completed_children_when_new_leading_child_completes_then_maintain_viewport_position_and_show_one_more_leading_completed_children() -> None:
    N = 5
    M = 3
    async with _project_with_resource_group_starting_to_download(
                resource_count=N + 4,
                small_max_visible_children=N,
                small_max_leading_complete_children=M,
                scheduler_thread_enabled=False,
            ) as (mw, download_rg_ttn, download_rg_members_ttn, _):
        project = Project._last_opened_project
        assert project is not None
        
        parent_ttn = download_rg_members_ttn
        
        # Ensure starts with no leading completed children
        assert 0 == _viewport_offset(parent_ttn)
        assert 0 + N == _materialized_child_task_count(parent_ttn)
        assert 0 == _unmaterialized_child_task_count(parent_ttn)
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == N + 1
        assert all([_is_download_resource_node(tn) for tn in children_tns[:-1]])
        assert _is_more_node(children_tns[-1])
        assert (
            [False] * N == 
            [_is_complete(tn) for tn in children_tns[:-1]]
        )
        
        # Case: test_given_showing_less_than_5_leading_completed_children_when_new_leading_child_completes_then_maintain_viewport_position_and_show_one_more_leading_completed_children
        _mark_as_complete(children_tasks[0])
        assert 0 == _viewport_offset(parent_ttn)
        assert 0 + N == _materialized_child_task_count(parent_ttn)
        assert 0 == _unmaterialized_child_task_count(parent_ttn)
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == N + 1
        assert all([_is_download_resource_node(tn) for tn in children_tns[:-1]])
        assert _is_more_node(children_tns[-1])
        assert (
            [True] + ([False] * (N - 1)) == 
            [_is_complete(tn) for tn in children_tns[:-1]]
        )
        
        # Case: test_given_showing_less_than_5_leading_completed_children_when_new_nonleading_child_completes_then_maintain_viewport_position_and_show_same_number_of_leading_completed_children
        assert M >= 3
        _mark_as_complete(children_tasks[2])
        assert 0 == _viewport_offset(parent_ttn)
        assert 0 + N == _materialized_child_task_count(parent_ttn)
        assert 0 == _unmaterialized_child_task_count(parent_ttn)
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == N + 1
        assert all([_is_download_resource_node(tn) for tn in children_tns[:-1]])
        assert _is_more_node(children_tns[-1])
        assert (
            [True, False, True] + ([False] * (N - 3)) == 
            [_is_complete(tn) for tn in children_tns[:-1]]
        )
        
        # Complete first M children
        assert N >= M + 1
        for t in children_tasks[:M]:
            if not t.complete:
                _mark_as_complete(t)
        assert 0 == _viewport_offset(parent_ttn)
        assert 0 + N == _materialized_child_task_count(parent_ttn)
        assert 0 == _unmaterialized_child_task_count(parent_ttn)
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == N + 1
        assert all([_is_download_resource_node(tn) for tn in children_tns[:-1]])
        assert _is_more_node(children_tns[-1])
        assert (
            ([True] * M) + ([False] * (N - M)) == 
            [_is_complete(tn) for tn in children_tns[:-1]]
        )
        
        # Case: test_given_showing_5_leading_completed_children_when_new_leading_child_completes_with_a_following_noncompleted_child_then_shift_viewport_down_once_and_still_show_5_leading_completed_children
        _mark_as_complete(children_tasks[M])
        assert 1 == _viewport_offset(parent_ttn)
        assert 1 + N == _materialized_child_task_count(parent_ttn)
        assert 1 == _unmaterialized_child_task_count(parent_ttn)
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == 1 + N + 1
        assert _is_more_node(children_tns[0])
        assert all([_is_download_resource_node(tn) for tn in children_tns[1:-1]])
        assert _is_more_node(children_tns[-1])
        assert (
            ([True] * M) + ([False] * (N - M)) == 
            [_is_complete(tn) for tn in children_tns[1:-1]]
        )
        
        # Case: test_given_showing_5_leading_completed_children_when_new_nonleading_child_completes_then_maintain_viewport_position_and_still_show_5_leading_completed_children
        assert N >= M + 2
        _mark_as_complete(children_tasks[M + 2])
        assert 1 == _viewport_offset(parent_ttn)
        assert 1 + N == _materialized_child_task_count(parent_ttn)
        assert 1 == _unmaterialized_child_task_count(parent_ttn)
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == 1 + N + 1
        assert _is_more_node(children_tns[0])
        assert all([_is_download_resource_node(tn) for tn in children_tns[1:-1]])
        assert _is_more_node(children_tns[-1])
        assert (
            ([True] * M) + [False, True] + ([False] * (N - (M + 2))) == 
            [_is_complete(tn) for tn in children_tns[1:-1]]
        )
        
        # Case: test_given_showing_5_leading_completed_children_when_new_leading_child_completes_with_a_following_completed_child_then_shift_viewport_down_multiple_times_and_still_show_5_leading_completed_children
        _mark_as_complete(children_tasks[M + 1])
        assert 3 == _viewport_offset(parent_ttn)
        assert 3 + N == _materialized_child_task_count(parent_ttn)
        assert 3 == _unmaterialized_child_task_count(parent_ttn)
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == 1 + N + 1
        assert _is_more_node(children_tns[0])
        assert all([_is_download_resource_node(tn) for tn in children_tns[1:-1]])
        assert _is_more_node(children_tns[-1])
        assert (
            ([True] * M) + ([False] * (N - M)) == 
            [_is_complete(tn) for tn in children_tns[1:-1]]
        )


@skip('covered by: test_given_showing_less_than_5_leading_completed_children_when_new_leading_child_completes_then_maintain_viewport_position_and_show_one_more_leading_completed_children')
async def test_given_showing_less_than_5_leading_completed_children_when_new_nonleading_child_completes_then_maintain_viewport_position_and_show_same_number_of_leading_completed_children() -> None:
    pass

    
@skip('covered by: test_given_showing_less_than_5_leading_completed_children_when_new_leading_child_completes_then_maintain_viewport_position_and_show_one_more_leading_completed_children')
async def test_given_showing_5_leading_completed_children_when_new_leading_child_completes_with_a_following_noncompleted_child_then_shift_viewport_down_once_and_still_show_5_leading_completed_children() -> None:
    pass

    
@skip('covered by: test_given_showing_less_than_5_leading_completed_children_when_new_leading_child_completes_then_maintain_viewport_position_and_show_one_more_leading_completed_children')
async def test_given_showing_5_leading_completed_children_when_new_nonleading_child_completes_then_maintain_viewport_position_and_still_show_5_leading_completed_children() -> None:
    pass

    
@skip('covered by: test_given_showing_less_than_5_leading_completed_children_when_new_leading_child_completes_then_maintain_viewport_position_and_show_one_more_leading_completed_children')
async def test_given_showing_5_leading_completed_children_when_new_leading_child_completes_with_a_following_completed_materialized_child_then_shift_viewport_down_multiple_times_and_still_show_5_leading_completed_children() -> None:
    pass


async def test_given_showing_5_leading_completed_children_when_new_leading_child_completes_with_a_following_completed_unmaterialized_child_then_shift_viewport_down_multiple_times_and_still_show_5_leading_completed_children() -> None:
    N = 5
    M = 3
    
    with _children_marked_as_complete_upon_creation([N + 1, N + 2]):
        async with _project_with_resource_group_starting_to_download(
                    resource_count=N + 4,
                    small_max_visible_children=N,
                    small_max_leading_complete_children=M,
                    scheduler_thread_enabled=False,
                ) as (mw, download_rg_ttn, download_rg_members_ttn, _):
            project = Project._last_opened_project
            assert project is not None
            
            parent_ttn = download_rg_members_ttn
            
            (children_tns, children_tasks) = \
                (parent_ttn.tree_node.children, parent_ttn.task.children)
            
            # Complete first (N-1) children
            assert N >= M + 1
            for t in children_tasks[:N-1]:
                if not t.complete:
                    _mark_as_complete(t)
            assert 1 == _viewport_offset(parent_ttn)
            (children_tns, children_tasks) = \
                (parent_ttn.tree_node.children, parent_ttn.task.children)
            assert len(children_tns) == N + 2
            assert all([_is_download_resource_node(tn) for tn in children_tns[1:-1]])
            assert 1 == _value_of_more_node(children_tns[0])
            assert 3 == _value_of_more_node(children_tns[-1])
            
            # Complete child N.
            # 
            # Ensure shifts viewport 3 times:
            # - once for completed child (N) and
            # - once for completed child (N + 1) that becomes materialized
            # - once for completed child (N + 2) that becomes materialized
            _mark_as_complete(children_tasks[N - 1])
            assert 4 == _viewport_offset(parent_ttn)
            (children_tns, children_tasks) = \
                (parent_ttn.tree_node.children, parent_ttn.task.children)
            assert len(children_tns) == N + 1
            assert all([_is_download_resource_node(tn) for tn in children_tns[1:]])
            assert 4 == _value_of_more_node(children_tns[0])
            #assert 0 == _value_of_more_node(children_tns[-1])


@contextmanager
def _children_marked_as_complete_upon_creation(ordinals: List[int]) -> Iterator[None]:
    # Prepare: Mark appropriate children as complete immediately upon creation
    super_goc_download_task = Resource.get_or_create_download_task
    def get_or_create_download_task(*args, **kwargs) -> 'Tuple[DownloadResourceTask, bool]':
        (task, created) = super_goc_download_task(*args, **kwargs)
        
        for n in ordinals:
            if task._resource.url.endswith(f'/https/xkcd.com/{n}/'):
                _mark_as_complete(task)
                break
        
        return (task, created)
    
    with patch('crystal.model.Resource.get_or_create_download_task', get_or_create_download_task):
        yield


# === Test: DownloadResourceGroupMembersTask: Observe when group member added ===

async def test_given_downloading_group_and_many_uncompleted_children_remaining_when_group_member_added_then_trailing_more_node_shows_download_task_for_member_added() -> None:
    N = 3
    M = 1
    async with _project_with_resource_group_starting_to_download(
                resource_count=N + 2,
                small_max_visible_children=N,
                small_max_leading_complete_children=M,
                scheduler_thread_enabled=False,
            ) as (mw, download_rg_ttn, download_rg_members_ttn, create_resource):
        project = Project._last_opened_project
        assert project is not None
        
        parent_ttn = download_rg_members_ttn
        
        # Ensure initially has trailing more node
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == N + 1
        assert all([_is_download_resource_node(tn) for tn in children_tns[:-1]])
        assert _is_more_node(children_tns[-1])
        assert 2 == _value_of_more_node(children_tns[-1])
        assert 0 + N == _materialized_child_task_count(parent_ttn)
        assert 0 == _unmaterialized_child_task_count(parent_ttn)
        
        # Case: test_given_downloading_group_and_many_uncompleted_children_remaining_when_group_member_added_then_trailing_more_node_shows_download_task_for_member_added
        create_resource()
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == N + 1
        assert all([_is_download_resource_node(tn) for tn in children_tns[:-1]])
        assert _is_more_node(children_tns[-1])
        assert 3 == _value_of_more_node(children_tns[-1])
        assert 0 + N == _materialized_child_task_count(parent_ttn)
        assert 0 == _unmaterialized_child_task_count(parent_ttn)
        
        # Complete tasks until more node disappears,
        # but adding a task will cause it to reappear
        assert N >= M+2
        for t in parent_ttn.task.children[:M+3]:
            _mark_as_complete(t)
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == 1 + N
        assert _is_more_node(children_tns[0])
        assert all([_is_download_resource_node(tn) for tn in children_tns[1:]])
        assert (M+2) + N == _materialized_child_task_count(parent_ttn)
        assert (M+2) == _unmaterialized_child_task_count(parent_ttn)
        
        # Case: test_given_downloading_group_and_few_uncompleted_children_remaining_when_group_member_added_then_trailing_more_node_added_with_value_1
        create_resource()
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == 1 + N + 1
        assert _is_more_node(children_tns[0])
        assert all([_is_download_resource_node(tn) for tn in children_tns[1:-1]])
        assert _is_more_node(children_tns[-1])
        assert 1 == _value_of_more_node(children_tns[-1])
        assert (M+2) + N == _materialized_child_task_count(parent_ttn)
        assert (M+2) == _unmaterialized_child_task_count(parent_ttn)
        
        # Complete tasks until more node disappears,
        # but adding a task will NOT cause it to reappear
        for t in parent_ttn.task.children[M+3:(M+3)+2]:
            _mark_as_complete(t)
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == 1 + (N - 1)
        assert _is_more_node(children_tns[0])
        assert all([_is_download_resource_node(tn) for tn in children_tns[1:]])
        assert ((M+2)+2) + (N - 1) == _materialized_child_task_count(parent_ttn)
        assert ((M+2)+2) == _unmaterialized_child_task_count(parent_ttn)
        
        # Case: test_given_downloading_group_and_few_uncompleted_children_remaining_when_group_member_added_then_download_task_for_member_added
        create_resource()
        (children_tns, children_tasks) = \
            (parent_ttn.tree_node.children, parent_ttn.task.children)
        assert len(children_tns) == 1 + N
        assert _is_more_node(children_tns[0])
        assert all([_is_download_resource_node(tn) for tn in children_tns[1:]])
        assert ((M+2)+2) + N == _materialized_child_task_count(parent_ttn)
        assert ((M+2)+2) == _unmaterialized_child_task_count(parent_ttn)


@skip('covered by: test_given_downloading_group_and_many_uncompleted_children_remaining_when_group_member_added_then_trailing_more_node_shows_download_task_for_member_added')
async def test_given_downloading_group_and_few_uncompleted_children_remaining_when_group_member_added_then_download_task_for_member_added() -> None:
    pass


@skip('covered by: test_given_downloading_group_and_many_uncompleted_children_remaining_when_group_member_added_then_trailing_more_node_shows_download_task_for_member_added')
async def test_given_downloading_group_and_few_uncompleted_children_remaining_when_group_member_added_then_trailing_more_node_added_with_value_1() -> None:
    pass


# === Test: RootTask ===

@skip('not yet automated')
async def test_when_top_level_task_finishes_then_is_removed_from_ui_soon(self) -> None:
    pass


# === Utility ===

@asynccontextmanager
async def _project_with_resource_group_starting_to_download(
        *, resource_count: int,
        small_max_visible_children: int,
        small_max_leading_complete_children: int,
        scheduler_thread_enabled: bool=True,
        ) -> AsyncIterator[Tuple[MainWindow, TaskTreeNode, TaskTreeNode, Callable[[], Resource]]]:
    if not (small_max_leading_complete_children <= small_max_visible_children):
        raise ValueError()
    
    scheduler_maybe_disabled = (
        scheduler_disabled()
        if not scheduler_thread_enabled
        else cast(AbstractContextManager, nullcontext())
    )  # type: AbstractContextManager
    
    # NOTE: NOT using the xkcd test data, because I want all xkcd URLs to give HTTP 404
    with scheduler_maybe_disabled, \
            served_project('testdata_bongo.cat.crystalproj.zip') as sp:
        # Define URLs
        if True:
            home_url = sp.get_request_url('https://xkcd.com/')
            
            comic_pattern = sp.get_request_url('https://xkcd.com/#/')
        
        # Use smaller values for children counts so that the test runs faster
        assert 100 == TaskTreeNode._MAX_VISIBLE_CHILDREN
        assert 5 == TaskTreeNode._MAX_LEADING_COMPLETE_CHILDREN
        with patch.object(TaskTreeNode, '_MAX_VISIBLE_CHILDREN', small_max_visible_children), \
                patch.object(TaskTreeNode, '_MAX_LEADING_COMPLETE_CHILDREN', small_max_leading_complete_children):
            
            async with (await OpenOrCreateDialog.wait_for()).create() as (mw, _):
                project = Project._last_opened_project
                assert project is not None
                
                # Create group
                g = ResourceGroup(project, 'Comic', comic_pattern)
                
                # Create group members
                for i in range(1, resource_count + 1):
                    Resource(project, comic_pattern.replace('#', str(i)))
                
                # Start downloading group
                assert 0 == len(project.root_task.children)
                g.download(); append_deferred_top_level_tasks(project)
                
                (download_rg_task,) = project.root_task.children
                assert isinstance(download_rg_task, DownloadResourceGroupTask)
                load_children_of_drg_task(download_rg_task, scheduler_thread_enabled=scheduler_thread_enabled)
                (_, download_rg_members_task) = download_rg_task.children
                assert isinstance(download_rg_members_task, DownloadResourceGroupMembersTask)
                
                # Define: create_resource()
                next_resource_ordinal = resource_count + 1
                def create_resource() -> Resource:
                    nonlocal next_resource_ordinal
                    r = Resource(project, comic_pattern.replace('#', str(next_resource_ordinal)))
                    next_resource_ordinal += 1
                    return r
                
                download_rg_ttn = _ttn_for_task(download_rg_task)
                download_rg_members_ttn = _ttn_for_task(download_rg_members_task)
                try:
                    yield (
                        mw,
                        download_rg_ttn,
                        download_rg_members_ttn,
                        create_resource,
                    )
                finally:
                    # Cleanup: Complete all children
                    _cleanup_download_of_resource_group(
                        download_rg_ttn,
                        download_rg_members_ttn,
                        project)


def _cleanup_download_of_resource_group(
        download_rg_ttn: TaskTreeNode,
        download_rg_members_ttn: TaskTreeNode,
        project: Project) -> None:
    parent_ttn = download_rg_members_ttn
    
    # Cleanup: Complete all children
    children = parent_ttn.task.children
    materialized_children = (
        [
            t for t in children._cached_prefix
            if not isinstance(t, UnmaterializedItem)
        ] + list(children[children.cached_prefix_len:])
        if isinstance(children, AppendableLazySequence)
        else children
    )
    for t in materialized_children:
        if not t.complete:
            _mark_as_complete(t)
    assert _is_complete(download_rg_members_ttn.tree_node)
    assert _is_complete(download_rg_ttn.tree_node)
    assert None == project.root_task.try_get_next_task_unit()  # clear root task
    assert len(project.root_task.children) == 0


def _viewport(ttn: TaskTreeNode) -> Tuple[int, int, bool, bool]:
    """
    Returns information about the visible tasks in the specified TaskTreeNode.
    """
    tn_children = ttn.tree_node.children
    if len(tn_children) == 0:
        return (0, 0, False, False)
    
    # Calculate: offset, has_first_more_node
    first_node = tn_children[0]
    if isinstance(first_node, _MoreNodeView):
        offset = first_node.more_count
        has_first_more_node = True
    else:
        offset = 0
        has_first_more_node = False
    
    # Calculate: length, has_last_more_node
    last_node = tn_children[-1]
    has_last_more_node = isinstance(last_node, _MoreNodeView)
    length = (
        len(tn_children) -
        (1 if has_first_more_node else 0) -
        (1 if has_last_more_node else 0)
    )
    
    return (offset, length, has_first_more_node, has_last_more_node)


def _viewport_offset(ttn: TaskTreeNode) -> int:
    (offset, _, _, _) = _viewport(ttn)
    return offset


def _viewport_length(ttn: TaskTreeNode) -> int:
    (_, length, _, _) = _viewport(ttn)
    return length


def _materialized_child_task_count(parent_ttn: TaskTreeNode) -> int:
    """
    Returns the number of task children that have ever been materialized,
    including any task children that were unmaterialized later.
    """
    children = parent_ttn.task.children
    if isinstance(children, AppendableLazySequence):
        return children.cached_prefix_len
    elif isinstance(children, list):
        return len(children)
    else:
        raise AssertionError(f'Unrecognized type of children list: {children!r}')


def _unmaterialized_child_task_count(parent_ttn: TaskTreeNode) -> int:
    """
    Returns the number of task children that were once materialized
    but were unmaterialized later.
    """
    children = parent_ttn.task.children
    if isinstance(children, AppendableLazySequence):
        return sum([
            isinstance(t, UnmaterializedItem)
            for t in children._cached_prefix
        ])
    elif isinstance(children, list):
        return 0
    else:
        raise AssertionError(f'Unrecognized type of children list: {children!r}')


def _is_download_resource_node(tn: NodeView) -> bool:
    return not _is_more_node(tn)


def _is_more_node(tn: NodeView) -> bool:
    return _value_of_more_node(tn) is not None


def _value_of_more_node(tn: NodeView) -> Optional[int]:
    tn_title = tn.title
    if not tn_title.endswith(' more'):
        return None
    return int(tn_title[:-len(' more')])


def _is_complete(tn: NodeView) -> bool:
    assert isinstance(tn, NodeView2)
    return tn.subtitle == 'Complete'



from contextlib import contextmanager
from crystal.tests import (
    test_bulkheads,
    test_disk_io_errors,
    test_do_not_download_groups,
    test_download,
    test_download_body,
    test_edit_group,
    test_edit_root_url,
    test_entitytree,
    test_file_extension_visibility,
    test_icons,
    test_install_to_desktop,
    test_load_urls,
    test_log_drawer,
    test_menus,
    test_new_group,
    test_new_root_url,
    test_open_project,
    test_parse_html,
    test_profile,
    test_project_migrate,
    test_readonly_mode,
    test_server,
    test_shell,
    test_ssd,
    test_tasks,
    test_tasktree,
    test_workflows,
    test_xthreading,
)
from crystal.tests.util.downloads import delay_between_downloads_minimized
from crystal.tests.util.runner import run_test
from crystal.tests.util.subtests import SubtestFailed
from crystal.util.xthreading import bg_affinity
from crystal.util.xtraceback import _CRYSTAL_PACKAGE_PARENT_DIRPATH
from crystal.util.xtime import sleep_profiled
from functools import wraps
import gc
import os
import sys
import time
import traceback
from typing import Any, Callable, Coroutine, Dict, List, Iterator, Optional, Tuple
from unittest import SkipTest
import warnings


def _test_functions_in_module(mod) -> List[Callable]:
    return [
        f for f in mod.__dict__.values() 
        if (
            callable(f) and 
            getattr(f, '__name__', '').startswith('test_') and
            # NOTE: Need to check stringness explicitly to exclude "call" from 
            #       "from unittest.mock import call"
            isinstance(getattr(f, '__name__', ''), str)
        )
    ]

# TODO: Avoid the need to manually enumerate all test modules individually
_TEST_FUNCS = (
    _test_functions_in_module(test_bulkheads) +
    _test_functions_in_module(test_disk_io_errors) +
    _test_functions_in_module(test_do_not_download_groups) +
    _test_functions_in_module(test_download) +
    _test_functions_in_module(test_download_body) +
    _test_functions_in_module(test_edit_group) +
    _test_functions_in_module(test_edit_root_url) +
    _test_functions_in_module(test_entitytree) +
    _test_functions_in_module(test_file_extension_visibility) +
    _test_functions_in_module(test_icons) +
    _test_functions_in_module(test_install_to_desktop) +
    _test_functions_in_module(test_load_urls) +
    _test_functions_in_module(test_log_drawer) +
    _test_functions_in_module(test_menus) +
    _test_functions_in_module(test_new_group) +
    _test_functions_in_module(test_new_root_url) +
    _test_functions_in_module(test_open_project) +
    _test_functions_in_module(test_parse_html) +
    _test_functions_in_module(test_profile) +
    _test_functions_in_module(test_project_migrate) +
    _test_functions_in_module(test_readonly_mode) +
    _test_functions_in_module(test_server) +
    _test_functions_in_module(test_shell) +
    _test_functions_in_module(test_ssd) +
    _test_functions_in_module(test_tasks) +
    _test_functions_in_module(test_tasktree) +
    _test_functions_in_module(test_workflows) +
    _test_functions_in_module(test_xthreading) +
    []
)


_TestFuncId = Tuple[str, str]  # (module, func_name)


# TODO: Rename "test_names" to something more appropriate,
#       now that items can also refer to test modules (and not just test functions)
@bg_affinity
def run_tests(test_names: List[str]) -> bool:
    """
    Runs automated UI tests, printing a summary report,
    and returning whether the run was OK.
    
    The format of the summary report is designed to be similar
    to that used by Python's unittest module.
    """
    with delay_between_downloads_minimized(), sleep_profiled():
        return _run_tests(test_names)


def _run_tests(test_names: List[str]) -> bool:
    assert os.environ.get('CRYSTAL_RUNNING_TESTS') == 'True'
    
    result_for_test_func_id = {}  # type: Dict[_TestFuncId, Optional[Exception]]
    start_time = time.time()  # capture
    run_count = 0
    with warnings.catch_warnings(record=True) as warning_list, _warnings_sent_to_ci():
        assert warning_list is not None
        
        for test_func in _TEST_FUNCS:
            if not callable(test_func):
                raise ValueError(f'Test function is not callable: {test_func}')
            test_func_id = (test_func.__module__, test_func.__name__)  # type: _TestFuncId
            test_name = f'{test_func_id[0]}.{test_func_id[1]}'
            
            # Only run test if it was requested (or if all tests are to be run)
            if len(test_names) > 0:
                if test_name not in test_names and test_func.__module__ not in test_names:
                    continue
            run_count += 1
            
            os.environ['CRYSTAL_SCREENSHOT_ID'] = test_name
            
            print('=' * 70)
            print(f'RUNNING: {test_func_id[1]} ({test_func_id[0]})')
            print('-' * 70)
            try:
                try:
                    run_test(test_func)
                finally:
                    # Flush any stderr output immediately,
                    # to keep aligned with stdout output from the test
                    sys.stderr.flush()
            except AssertionError as e:
                result_for_test_func_id[test_func_id] = e
                
                traceback.print_exc(file=sys.stdout)
                print('FAILURE')
            except SkipTest as e:
                result_for_test_func_id[test_func_id] = e
                
                print(f'SKIP ({str(e)})')
            except Exception as e:
                result_for_test_func_id[test_func_id] = e
                
                if not isinstance(e, SubtestFailed):
                    traceback.print_exc(file=sys.stdout)
                print(f'ERROR ({e.__class__.__name__})')
            else:
                result_for_test_func_id[test_func_id] = None
                
                print('OK')
            print()
            
            # Garbage collect, running any finalizers in __del__() early,
            # such as the warnings printed by ListenableMixin
            gc.collect()
    
    end_time = time.time()  # capture
    delta_time = end_time - start_time
    
    failure_count = 0
    error_count = 0
    skip_count = 0
    failed_test_names = []
    for (test_func_id, result) in result_for_test_func_id.items():
        if result is None:
            pass
        elif isinstance(result, SkipTest):
            skip_count += 1
        else:
            if isinstance(result, AssertionError):
                failure_count += 1
            else:
                error_count += 1
            
            test_name = f'{test_func_id[0]}.{test_func_id[1]}'
            failed_test_names.append(test_name)
    
    is_ok = (failure_count + error_count) == 0
    
    suffix_parts = []
    if failure_count != 0:
        suffix_parts.append(f'failures={failure_count}')
    if error_count != 0:
        suffix_parts.append(f'errors={error_count}')
    if skip_count != 0:
        suffix_parts.append(f'skipped={skip_count}')
    suffix = (
        f' ({", ".join(suffix_parts)})'
        if len(suffix_parts) != 0
        else ''
    )
    
    print('=' * 70)
    print('SUMMARY')
    print('-' * 70)
    for result in result_for_test_func_id.values():
        if result is None:
            print('.', end='')
        elif isinstance(result, AssertionError):
            print('F', end='')
        elif isinstance(result, SkipTest):
            if str(result).startswith('covered by:'):
                print('c', end='')
            else:
                print('s', end='')
        else:
            print('E', end='')
    print()
    
    print('-' * 70)
    print(f'Ran {run_count} tests in {"%.3f" % delta_time}s')
    print()
    print(f'{"OK" if is_ok else "FAILURE"}{suffix}')
    
    # Print warnings, if any
    if len(warning_list) >= 1:
        print()
        print('Warnings:')
        for w in warning_list:
            if w.filename.startswith(_CRYSTAL_PACKAGE_PARENT_DIRPATH):
                short_filepath = os.path.relpath(w.filename, start=_CRYSTAL_PACKAGE_PARENT_DIRPATH)
            else:
                short_filepath = w.filename
            w_str = warnings.formatwarning(w.message, w.category, short_filepath, w.lineno, w.line)
            print('- ' + w_str, end='')
    
    # Print command to rerun failed tests
    if len(failed_test_names) != 0:
        print()
        print('Rerun failed tests with:')
        print(f'$ crystal --test {" ".join(failed_test_names)}')
        print()
    
    # Play bell sound in terminal
    print('\a', end='', flush=True)
    
    return is_ok


@contextmanager
def _warnings_sent_to_ci() -> Iterator[None]:
    if not _running_in_ci():
        yield
        return
    
    super_showwarning = warnings.showwarning  # capture
    
    def showwarning(message, category, filename, lineno, file=None, line=None):
        # Try to reformat `filename` to use the Linux format so that warning
        # annotations are associated with the correct file
        # 
        # Depending on OS, `filename` initially looks like:
        #     - macOS: setup/dist/Crystal Web Archiver.app/Contents/Resources/lib/python38.zip/crystal/tests/index.py
        #     - Linux: src/crystal/tests/index.py
        #     - Windows: crystal\tests\index.pyc
        filename_parts = filename.split(os.path.sep)
        if 'crystal' in filename_parts:
            filename_parts = ['src'] + filename_parts[filename_parts.index('crystal'):]
            if filename_parts[-1].endswith('.pyc'):
                filename_parts[-1] = filename_parts[-1][:-1]  # convert .pyc ending to .py
            filename = '/'.join(filename_parts)  # reinterpret
        
        # Create warning annotation in GitHub Action's [Summary > Annotations] section
        # 
        # Syntax: https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#example-setting-a-warning-message
        print(f'::warning file={filename},line={lineno}::{message}')
        
        return super_showwarning(message, category, filename, lineno, file, line)
    
    warnings.showwarning = showwarning
    try:
        yield
    finally:
        warnings.showwarning = super_showwarning


def _running_in_ci() -> bool:
    return os.environ.get('GITHUB_ACTIONS') == 'true'



import io
import os.path
import traceback


_CRYSTAL_PACKAGE_PARENT_DIRPATH = os.path.abspath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
_PYTHON_STDLIB_PARENT_DIRPATH = os.path.abspath(os.path.dirname(os.__file__))


def format_exception_for_user(exc: BaseException) -> str:
    tb = exc.__traceback__
    
    exception_str = io.StringIO()
    print(f'{type(exc).__name__}: {str(exc)}', file=exception_str)
    for fs in reversed(traceback.extract_tb(tb)):
        if '# cr-traceback: ignore' in (fs.line or ''):
            # Don't print frames marked as uninteresting
            continue
        if fs.filename.startswith(_CRYSTAL_PACKAGE_PARENT_DIRPATH):
            short_filepath = os.path.relpath(fs.filename, start=_CRYSTAL_PACKAGE_PARENT_DIRPATH)
        elif fs.filename.startswith(_PYTHON_STDLIB_PARENT_DIRPATH):
            short_filepath = os.path.relpath(fs.filename, start=_PYTHON_STDLIB_PARENT_DIRPATH)
        else:
            short_filepath = fs.filename
        print(f'  at {short_filepath}:{fs.lineno} in {fs.name}', file=exception_str)
    
    return exception_str.getvalue()



from crystal.task import (
    ASSUME_RESOURCES_DOWNLOADED_IN_SESSION_WILL_ALWAYS_REMAIN_FRESH,
    ProjectFreeSpaceTooLowError, Task,
)
from crystal.tests.util.asserts import *
from crystal.tests.util.data import (
    MAX_TIME_TO_DOWNLOAD_404_URL,
    MAX_TIME_TO_DOWNLOAD_XKCD_HOME_URL_BODY
)
from crystal.tests.util.downloads import load_children_of_drg_task
from crystal.tests.util.server import served_project
from crystal.tests.util.skip import skipTest
from crystal.tests.util.subtests import SubtestsContext, awith_subtests
from crystal.tests.util.tasks import scheduler_thread_context
from crystal.tests.util.wait import wait_for
from crystal.tests.util.windows import OpenOrCreateDialog
from crystal.model import Project, Resource, ResourceGroup
from crystal.util.progress import ProgressBarCalculator
from crystal.util.xcollections.lazy import AppendableLazySequence
from typing import NamedTuple
from unittest import skip
from unittest.mock import patch, Mock, PropertyMock


# ==============================================================================
# Test: Tasks

# ------------------------------------------------------------------------------
# Test: Task

@awith_subtests
async def test_some_tasks_may_complete_immediately(subtests) -> None:
    assert True == ASSUME_RESOURCES_DOWNLOADED_IN_SESSION_WILL_ALWAYS_REMAIN_FRESH, \
        'Expected optimization to be enabled: ASSUME_RESOURCES_DOWNLOADED_IN_SESSION_WILL_ALWAYS_REMAIN_FRESH'
    
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        if True:
            missing_url = sp.get_request_url('https://example.com/')
            
            comic_pattern = sp.get_request_url('https://xkcd.com/#/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, _):
            project = Project._last_opened_project
            assert project is not None
            
            missing_r = Resource(project, missing_url)
            
            with subtests.test(task_type='DownloadResourceTask'):
                # Download the resource
                assert False == missing_r.already_downloaded_this_session
                missing_rr_future = missing_r.download()  # uses DownloadResourceTask
                await wait_for(
                    lambda: missing_rr_future.done() or None,
                    timeout=MAX_TIME_TO_DOWNLOAD_404_URL)
                assert True == missing_r.already_downloaded_this_session
                
                # Download the resource again, and ensure it downloads immediately
                dr_task = missing_r.create_download_task(needs_result=False)  # a DownloadResourceTask
                assert True == dr_task.complete
            
            with subtests.test(task_type='UpdateResourceGroupMembersTask'):
                # Covered by subtest: DownloadResourceGroupTask
                pass
            
            with subtests.test(task_type='DownloadResourceGroupMembersTask'):
                # Covered by subtest: DownloadResourceGroupTask
                pass
            
            with subtests.test(task_type='DownloadResourceGroupTask'):
                comic_rs = [
                    Resource(project, comic_pattern.replace('#', str(ordinal)))
                    for ordinal in [1, 2]
                ]
                
                comic_g = ResourceGroup(project, 'Comic', comic_pattern)
                assert None == comic_g.source
                
                COMIC_G_FINAL_MEMBER_COUNT = 10
                
                # Download the group (and all of its currently known members)
                for r in comic_rs:
                    assert False == r.already_downloaded_this_session
                drg_task = comic_g.create_download_task()  # a DownloadResourceGroupTask
                project.add_task(drg_task)
                await wait_for(
                    lambda: drg_task.complete or None,
                    timeout=(
                        MAX_TIME_TO_DOWNLOAD_404_URL +
                        (MAX_TIME_TO_DOWNLOAD_XKCD_HOME_URL_BODY * COMIC_G_FINAL_MEMBER_COUNT)
                    ))
                assert COMIC_G_FINAL_MEMBER_COUNT == len(comic_g.members)
                for r in comic_rs:
                    assert True == r.already_downloaded_this_session
                
                # Download the group again, and ensure it downloads immediately
                if True:
                    drg_task = comic_g.create_download_task()
                    
                    load_children_of_drg_task(drg_task, task_added_to_project=False)
                    
                    # Precondition
                    # NOTE: The group won't appear to be immediately downloaded yet
                    #       because no code has tried to access the lazily-loaded
                    #       DownloadResourceTask children yet and thus doesn't
                    #       know that all of those children are complete
                    assert (True, 0, False) == (
                        isinstance(drg_task._download_members_task.children, AppendableLazySequence),
                        drg_task._download_members_task.children.cached_prefix_len
                            if isinstance(drg_task._download_members_task.children, AppendableLazySequence)
                            else None,
                        drg_task.complete
                    )
                    
                    # TODO: Disable the scheduler thread before trying to control
                    #       it manually. Currently it's not safe to make the
                    #       scheduler_thread_context() assertion.
                    project.add_task(drg_task)
                    with scheduler_thread_context():
                        task_unit = project.root_task.try_get_next_task_unit()
                        assert task_unit is None
                    
                    # Postcondition
                    # NOTE: Adding the DownloadResourceGroupTask to the project's
                    #       task tree will cause the TaskTreeNode to start accessing
                    #       the DownloadResourceTask children because it wants to
                    #       create a paired TaskTreeNode for each such child.
                    #       These accesses cause the DownloadResourceTask children
                    #       to be created and observed as being already complete.
                    #       With all of those children complete the ancestor
                    #       DownloadResourceGroupTask will also be completed.
                    assert (True, COMIC_G_FINAL_MEMBER_COUNT, True) == (
                        isinstance(drg_task._download_members_task.children, AppendableLazySequence),
                        drg_task._download_members_task.children.cached_prefix_len
                            if isinstance(drg_task._download_members_task.children, AppendableLazySequence)
                            else None,
                        drg_task.complete
                    )


def test_given_running_tests_then_uses_extra_listener_assertions() -> None:
    assert True == Task._USE_EXTRA_LISTENER_ASSERTIONS_ALWAYS


# ------------------------------------------------------------------------------
# Test: DownloadResourceTask
#       (+ DownloadResourceBodyTask)
#       (+ ParseResourceRevisionLinks)

# (TODO: Add basic tests)

@skip('not yet automated')
async def test_when_download_resource_then_displays_estimated_time_remaining() -> None:
    pass


def test_format_of_estimated_time_remaining() -> None:
    if ProgressBarCalculator._VERBOSE:
        skipTest('cannot check format when verbose format is being used')
        return
    
    def format_remaining_str(second_count: int) -> str:
        class MockProgressBarCalculator(ProgressBarCalculator):
            pass
        
        pbc = MockProgressBarCalculator(initial=0, total=second_count)
        type(pbc).n = PropertyMock(return_value=0)  # type: ignore[assignment]
        type(pbc).total = PropertyMock(return_value=second_count)
        pbc._rc_n = Mock()
        type(pbc._rc_n).rate = PropertyMock(return_value=1)
        pbc._rc_total = Mock()
        type(pbc._rc_total).rate = PropertyMock(return_value=None)
        
        (remaining_str, time_per_item_str) = pbc.remaining_str_and_time_per_item_str()
        return remaining_str
    
    assertEqual('00:01', format_remaining_str(1))  # 1 second
    assertEqual('01:00', format_remaining_str(60))  # 1 minute
    assertEqual('1:00:00', format_remaining_str(60*60))  # 1 hr
    assertEqual('1d + 0:00:00', format_remaining_str(24*60*60))  # 1 day
    assertEqual('1d + 1:01:01', format_remaining_str(24*60*60 + 60*60 + 60 + 1))


# TODO: Extend test to check how user-facing UI responds to ProjectFreeSpaceTooLowError
#       condition. Currently the UI doesn't handle this condition well and
#       silently fails downloads.
@awith_subtests
async def test_given_project_on_disk_with_low_space_free_when_try_to_download_resource_revision_then_raises_exception(subtests: SubtestsContext) -> None:
    async def try_download_with_disk_usage(du: _DiskUsage, *, expect_failure: bool) -> None:
        assert project is not None
        with patch('shutil.disk_usage', return_value=du):
            r = Resource(project, home_url)
            try:
                rr_future = r.download_body()
                await wait_for(
                    lambda: rr_future.done() or None,
                    timeout=MAX_TIME_TO_DOWNLOAD_XKCD_HOME_URL_BODY)
                try:
                    rr = rr_future.result()
                except ProjectFreeSpaceTooLowError:
                    if expect_failure:
                        pass  # expected
                    else:
                        assert False, f'Expected ProjectFreeSpaceTooLowError but got success when {du}'
                else:
                    if not expect_failure:
                        pass  # expected
                    else:
                        assert False, f'Expected success but got ProjectFreeSpaceTooLowError when {du}'
            finally:
                r.delete()
    
    with subtests.test('given project on small disk and less than 5 percent of disk free'):
        with served_project('testdata_xkcd.crystalproj.zip') as sp:
            # Define URLs
            home_url = sp.get_request_url('https://xkcd.com/')
            
            async with (await OpenOrCreateDialog.wait_for()).create() as (mw, _):
                project = Project._last_opened_project
                assert project is not None
                
                await try_download_with_disk_usage(
                    _DiskUsage(total=100*1024*1024, used=(100-6)*1024*1024, free=6*1024*1024),
                    expect_failure=False)
                
                await try_download_with_disk_usage(
                    _DiskUsage(total=100*1024*1024, used=(100-4)*1024*1024, free=4*1024*1024),
                    expect_failure=True)
    
    with subtests.test('given project on large disk and less than 4 gib of disk free'):
        with served_project('testdata_xkcd.crystalproj.zip') as sp:
            # Define URLs
            home_url = sp.get_request_url('https://xkcd.com/')
            
            async with (await OpenOrCreateDialog.wait_for()).create() as (mw, _):
                project = Project._last_opened_project
                assert project is not None
                
                await try_download_with_disk_usage(
                    _DiskUsage(total=1000*1024*1024*1024, used=(1000-6)*1024*1024*1024, free=6*1024*1024*1024),
                    expect_failure=False)
                
                await try_download_with_disk_usage(
                    _DiskUsage(total=1000*1024*1024*1024, used=(1000-3)*1024*1024*1024, free=3*1024*1024*1024),
                    expect_failure=True)


class _DiskUsage(NamedTuple):
    total: int
    used: int
    free: int


# ------------------------------------------------------------------------------
# Test: DownloadResourceGroupTask
#       (+ UpdateResourceGroupMembersTask)
#       (+ DownloadResourceGroupMembersTask)

# (TODO: Add basic tests)

@skip('not yet automated')
async def test_when_download_resource_group_members_then_displays_estimated_time_remaining() -> None:
    pass


# ------------------------------------------------------------------------------
# Test: RootTask

# (TODO: Add basic tests)


# ==============================================================================
# Test: Scheduler

@skip('covered by: test_when_scheduler_thread_event_loop_crashes_then_RT_marked_as_crashed_and_scheduler_crashed_task_appears')
async def test_when_scheduler_thread_crashes_then_scheduler_crashed_task_appears():
    pass


# ==============================================================================



"""
Tests that projects in older on-disk formats are upgraded to the latest
format properly.

In particular this module tests the Project._apply_migrations() method.

All tests below implicitly include the condition:
* given_project_opened_as_writable
"""
from contextlib import asynccontextmanager, contextmanager
import crystal.main
from crystal.model import (
    Project, Resource, ResourceRevision as RR, ResourceRevisionMetadata,
)
from crystal import progress
from crystal.progress import CancelOpenProject, OpenProjectProgressDialog
from crystal.tests.test_server import serve_and_fetch_xkcd_home_page
from crystal.tests.util.controls import click_button
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.server import extracted_project
from crystal.tests.util.wait import wait_for, window_condition
from crystal.tests.util.windows import MainWindow, OpenOrCreateDialog
from crystal.util.db import DatabaseCursor
import os.path
from typing import AsyncIterator, cast, Iterator, Optional, Tuple
from typing_extensions import override
from unittest import skip
from unittest.mock import patch
import wx


# === .crystalopen & README ===

@skip('not yet automated')
def test_given_project_lacks_crystalopen_and_lacks_readme_when_project_opened_then_crystalopen_and_readme_created() -> None:
    pass


@skip('not yet automated')
def test_given_project_lacks_crystalopen_and_has_maybe_modified_readme_when_project_opened_then_readme_content_preserved() -> None:
    pass


@skip('not yet automated')
def test_given_project_has_crystalopen_with_nondefault_name_when_project_opened_then_crystalopen_name_preserved() -> None:
    pass


@skip('not yet automated')
def test_given_project_has_crystalopen_and_has_maybe_modified_readme_when_project_opened_then_readme_content_preserved() -> None:
    pass


@skip('not yet automated')
def test_given_project_has_crystalopen_and_readme_was_deleted_when_project_opened_then_readme_stays_deleted() -> None:
    pass


# === Windows desktop.ini ===

@skip('not yet automated')
def test_given_project_lacks_desktop_ini_file_then_desktop_ini_and_icons_directory_created() -> None:
    pass


@skip('not yet automated')
def test_given_project_has_desktop_ini_file_then_desktop_ini_content_preserved() -> None:
    pass


# === Linux .directory ===

@skip('not yet automated')
def test_given_project_lacks_dot_directory_file_then_dot_directory_file_created() -> None:
    pass


@skip('not yet automated')
def test_given_project_has_dot_directory_file_then_dot_directory_file_content_preserved() -> None:
    pass


# === Hide special files on Windows ===

@skip('not yet automated')
def test_given_windows_then_desktop_ini_file_and_dot_directory_file_marked_as_hidden() -> None:
    pass


# === Major Version Migrations ===

# --- Unknown Major Versions ---

async def test_refuses_to_open_project_with_unknown_high_major_version() -> None:
    UNKNOWN_HIGH_MAJOR_VERSION = Project._LATEST_SUPPORTED_MAJOR_VERSION + 1
    
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with unknown high major version
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as mw:
            project = Project._last_opened_project
            assert project is not None
            
            project._set_property('major_version', str(UNKNOWN_HIGH_MAJOR_VERSION))
            assert UNKNOWN_HIGH_MAJOR_VERSION == project.major_version
        
        # Try to open that project
        if True:
            # Prepare to: OK
            did_respond_to_project_too_new_modal = False
            def click_ok_in_project_too_new_modal(dialog: wx.Dialog) -> int:
                assert 'cr-project-too-new' == dialog.Name
                
                nonlocal did_respond_to_project_too_new_modal
                did_respond_to_project_too_new_modal = True
                
                return wx.ID_OK
            
            load_project = crystal.main._load_project  # capture
            def patched_load_project(*args, **kwargs) -> Project:
                return load_project(
                    *args,
                    _show_modal_func=click_ok_in_project_too_new_modal,  # type: ignore[misc]
                    **kwargs)
            
            with patch(f'crystal.main._load_project', patched_load_project):
                ocd = await OpenOrCreateDialog.wait_for()
                await ocd.start_opening(project_dirpath, next_window_name='cr-open-or-create-project')
                
                # HACK: Wait minimum duration to allow open to finish
                await bg_sleep(0.5)
                
                # Wait for cancel and return to initial dialog
                ocd = await OpenOrCreateDialog.wait_for()
                
                assert did_respond_to_project_too_new_modal


# --- 1 -> 2: Prompt to upgrade ---

@skip('covered by: test_can_upgrade_project_from_major_version_1_to_2')
async def test_given_project_is_major_version_1_and_would_be_fast_to_upgrade_to_major_version_2_when_open_project_then_updates_project_without_prompting_for_confirmation() -> None:
    pass


async def test_when_prompted_to_upgrade_project_from_major_version_1_to_2_then_can_continue_upgrade() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with major version = 1 and at least one revision
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
        
        # Prepare to: Accept upgrade
        did_respond_to_upgrade_required_modal = False
        def click_continue_in_upgrade_required_modal(dialog: wx.Dialog) -> int:
            assert 'cr-upgrade-required' == dialog.Name
            
            nonlocal did_respond_to_upgrade_required_modal
            did_respond_to_upgrade_required_modal = True
            
            style = dialog.GetMessageDialogStyle()
            if style & wx.YES_NO != 0:
                return wx.ID_YES  # Continue
            elif style & wx.OK != 0:
                return wx.ID_OK  # Continue
            else:
                raise AssertionError()
        
        with _upgrade_required_modal_always_shown(), \
                patch('crystal.progress.ShowModal', click_continue_in_upgrade_required_modal):
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as mw:
                assert did_respond_to_upgrade_required_modal
                
                maybe_project = Project._last_opened_project
                assert maybe_project is not None
                project = maybe_project
                assert project.major_version >= 2, (
                    'Expected project to be upgraded'
                )  # not == 1


async def test_when_prompted_to_upgrade_project_from_major_version_1_to_2_then_can_defer_upgrade_to_later() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with major version = 1 and at least one revision
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
        
        # Prepare to: Defer upgrade
        did_respond_to_upgrade_required_modal = False
        def click_later_in_upgrade_required_modal(dialog: wx.Dialog) -> int:
            assert 'cr-upgrade-required' == dialog.Name
            
            nonlocal did_respond_to_upgrade_required_modal
            did_respond_to_upgrade_required_modal = True
            
            style = dialog.GetMessageDialogStyle()
            if style & wx.YES_NO != 0:
                return wx.ID_NO  # Later
            elif style & wx.OK != 0:
                raise AssertionError('"Later" button in dialog is not visible')
            else:
                raise AssertionError()
        
        with _upgrade_required_modal_always_shown(), \
                patch('crystal.progress.ShowModal', click_later_in_upgrade_required_modal):
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as mw:
                assert did_respond_to_upgrade_required_modal
                
                maybe_project = Project._last_opened_project
                assert maybe_project is not None
                project = maybe_project
                assert 1 == project.major_version, (
                    'Expected project to not be upgraded'
                )  # not > 1


async def test_when_prompted_to_upgrade_project_from_major_version_1_to_2_then_can_cancel_open_project() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with major version = 1 and at least one revision
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
        
        # Prepare to: Cancel
        did_respond_to_upgrade_required_modal = False
        def click_cancel_in_upgrade_required_modal(dialog: wx.Dialog) -> int:
            assert 'cr-upgrade-required' == dialog.Name
            
            nonlocal did_respond_to_upgrade_required_modal
            did_respond_to_upgrade_required_modal = True
            
            return wx.ID_CANCEL
        
        with _upgrade_required_modal_always_shown(), \
                patch('crystal.progress.ShowModal', click_cancel_in_upgrade_required_modal):
            ocd = await OpenOrCreateDialog.wait_for()
            await ocd.start_opening(project_dirpath, next_window_name='cr-open-or-create-project')
            
            # HACK: Wait minimum duration to allow open to finish
            await bg_sleep(0.5)
            
            # Wait for cancel and return to initial dialog
            ocd = await OpenOrCreateDialog.wait_for()


async def test_given_project_is_major_version_1_and_has_more_than_MAX_REVISION_ID_revisions_when_open_project_then_defers_upgrade_to_later() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
        
        # Open project, simulating an estimated revision count > Project._MAX_REVISION_ID.
        # Ensure is not upgraded to major version 2.
        if True:
            super_process_table_rows = Project._process_table_rows
            @staticmethod  # type: ignore[misc]
            def process_table_rows(
                    c: DatabaseCursor,
                    approx_row_count_query: str,
                    *args, **kwargs,
                    ) -> None:
                return super_process_table_rows(
                    c,
                    f'select {Project._MAX_REVISION_ID + 1}'
                        if approx_row_count_query.startswith('select id from resource_revision ')
                        else approx_row_count_query,
                    *args, **kwargs)
        
        with patch.object(Project, '_process_table_rows', process_table_rows):
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as mw:
                maybe_project = Project._last_opened_project
                assert maybe_project is not None
                project = maybe_project
                
                assert 1 == project.major_version, (
                    'Expected project to not be upgraded'
                )  # not > 1


# --- 1 -> 2: Move revisions from flat structure to hierarchy ---

async def test_can_upgrade_project_from_major_version_1_to_2() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with >=4,096 revisions, which is just over the number of
        # revisions that fit into a single subdirectory of the new hierarchy
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
            
            # Create many resources
            max_id = max([r._id for r in project.resources])
            resources = Resource.bulk_create(project, [
                f'https://example.com/{id}'
                for id in range(max_id + 1, 4096 + 1)
            ], 'https://example.com/')
            max_id = max([r._id for r in project.resources])
            assert 4096 == max_id
            
            # Create many revisions
            if True:
                # Create revision rows in database
                c = project._db.cursor()
                encoded_null_error = RR._encode_error(None)
                encoded_404_error = RR._encode_metadata(ResourceRevisionMetadata(
                    http_version=10,
                    status_code=404,
                    reason_phrase='Not Found',
                    headers=[],
                ))
                c.executemany(
                    'insert into resource_revision '
                        '(id, resource_id, request_cookie, error, metadata) values (?, ?, ?, ?, ?)', 
                    [
                        (r._id, r._id, None, encoded_null_error, encoded_404_error)
                        for r in resources
                    ])
                project._db.commit()
                
                # Create revision files in filesystem
                for id in [r._id for r in resources]:
                    with open(os.path.join(project.path, Project._REVISIONS_DIRNAME, str(id)), 'w') as tf:
                        tf.write(str(id))  # arbitrary content
        
        # Upgrade the project to major version >= 2.
        # Ensure revisions appear to be migrated correctly.
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as mw:
            maybe_project = Project._last_opened_project
            assert maybe_project is not None
            project = maybe_project
            
            assert project.major_version >= 2, (
                'Expected project to be upgraded'
            )  # not == 1
            
            assert os.path.isdir(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '000'))
            assert os.path.isfile(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '000', '001'))
            
            assert os.path.isdir(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '001'))
            assert os.path.isfile(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '001', '000'))
            assert not os.path.exists(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '001', '001'))
            
            assert not os.path.exists(os.path.join(
                project.path, Project._REVISIONS_DIRNAME, '000', '000', '000', '002'))


async def test_can_cancel_and_resume_upgrade_of_project_from_major_version_1_to_2() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        # Create project with major version = 1 and at least one revision
        async with _project_opened_without_migrating(project_dirpath) as (_, project):
            assert 1 == project.major_version, (
                'Expected project to not be upgraded'
            )  # not > 1
        old_revision_count = _count_files_in(
            os.path.join(project.path, Project._REVISIONS_DIRNAME))  # capture
        assert old_revision_count >= 1
        
        # Start upgrading the project to major version >= 2,
        # but cancel the upgrade after the second revision
        if True:
            ocd = await OpenOrCreateDialog.wait_for()
            
            progress_listener = progress._active_progress_listener
            assert progress_listener is not None
            
            # Prepare to: Cancel upgrade in the middle
            did_cancel_migration_in_middle = False  # type: bool
            def upgrading_revision(index: int, revisions_per_second: float) -> None:
                nonlocal did_cancel_migration_in_middle
                if index >= 1:
                    # Cancel before migrating the second revision
                    did_cancel_migration_in_middle = True
                    raise CancelOpenProject()
            
            with patch.object(progress_listener, 'upgrading_revision', upgrading_revision), \
                    _progress_reported_at_maximum_resolution():
                await ocd.start_opening(project_dirpath, next_window_name='cr-open-or-create-project')
                
                # HACK: Wait minimum duration to allow open to finish
                await bg_sleep(0.5)
                
                # Wait for migration to start, get cancelled, and return to initial dialog
                ocd = await OpenOrCreateDialog.wait_for()
                assert did_cancel_migration_in_middle
        
        # Verify looks like migration was canceled in filesystem
        assert os.path.exists(os.path.join(project.path, Project._REVISIONS_DIRNAME))
        assert os.path.exists(os.path.join(project.path, Project._IN_PROGRESS_REVISIONS_DIRNAME))
        assert 1 == _count_files_in(
            os.path.join(project.path, Project._IN_PROGRESS_REVISIONS_DIRNAME))
        assert (old_revision_count - 1) == _count_files_in(
            os.path.join(project.path, Project._REVISIONS_DIRNAME))
        
        # Resume upgrading the project to major version >= 2. Allow to finish.
        if True:
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as mw:
                maybe_project = Project._last_opened_project
                assert maybe_project is not None
                project = maybe_project
                
                assert project.major_version >= 2, (
                    'Expected project to be upgraded'
                )  # not == 1
            new_revision_count = _count_files_in(
                os.path.join(project.path, Project._REVISIONS_DIRNAME))
        
        # Ensure all revisions appear to be migrated
        assert os.path.exists(os.path.join(project.path, Project._REVISIONS_DIRNAME))
        assert not os.path.exists(os.path.join(project.path, Project._IN_PROGRESS_REVISIONS_DIRNAME))
        assert old_revision_count == new_revision_count
        assert not os.path.exists(os.path.join(project.path, Project._TEMPORARY_DIRNAME, Project._REVISIONS_DIRNAME))


# --- 1 -> 2: Smoke test operations on both project versions ---

async def test_can_serve_revisions_from_project_with_major_version_1() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        async with _project_opened_without_migrating(project_dirpath) as (mw, project):
            assert 1 == project.major_version
            
            (server_page, _) = await serve_and_fetch_xkcd_home_page(mw)
            assert 200 == server_page.status


async def test_can_serve_revisions_from_project_with_major_version_2() -> None:
    with extracted_project('testdata_xkcd.crystalproj.zip') as project_dirpath:
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as mw:
            maybe_project = Project._last_opened_project
            assert maybe_project is not None
            project = maybe_project
            
            assert 2 == project.major_version
            
            (server_page, _) = await serve_and_fetch_xkcd_home_page(mw)
            assert 200 == server_page.status


# --- Utility ---

@asynccontextmanager
async def _project_opened_without_migrating(
        project_dirpath: str
        ) -> AsyncIterator[Tuple[MainWindow, Project]]:
    class NonUpgradingProject(Project):
        @override
        def _apply_migrations(self, *args, **kwargs) -> None:
            pass
    
    with patch(
            # Project imported inside by _prompt_to_open_project()
            'crystal.model.Project',
            NonUpgradingProject):
        async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath) as mw:
            project = NonUpgradingProject._last_opened_project
            assert project is not None
            
            yield (mw, project)


@contextmanager
def _upgrade_required_modal_always_shown() -> Iterator[None]:
    was_enabled = OpenProjectProgressDialog._always_show_upgrade_required_modal
    OpenProjectProgressDialog._always_show_upgrade_required_modal = True
    try:
        yield
    finally:
        OpenProjectProgressDialog._always_show_upgrade_required_modal = was_enabled


@contextmanager
def _progress_reported_at_maximum_resolution() -> Iterator[None]:
    was_enabled = Project._report_progress_at_maximum_resolution
    Project._report_progress_at_maximum_resolution = True
    try:
        yield
    finally:
        Project._report_progress_at_maximum_resolution = was_enabled


def _count_files_in(dirpath: str) -> int:
    file_count = 0
    for (_, _, filenames) in os.walk(dirpath):
        file_count += len(filenames)
    return file_count



from contextlib import contextmanager
from typing import Callable, Iterator, NoReturn, Optional


def is_wrapped_object_deleted_error(e: Exception) -> bool:
    e_str = str(e)
    return (
        isinstance(e, RuntimeError) and
        e_str.startswith('wrapped C/C++ object of type ') and
        e_str.endswith(' has been deleted')
    )


@contextmanager
def wrapped_object_deleted_error_ignored() -> Iterator[None]:
    try:
        yield
    except Exception as e:
        if is_wrapped_object_deleted_error(e):
            pass
        else:
            raise


@contextmanager
def wrapped_object_deleted_error_raising(
        raiser: Optional[Callable[[], NoReturn]]=None
        ) -> Iterator[None]:
    if raiser is None:
        def default_raiser() -> NoReturn:
            raise WindowDeletedError()
        raiser = default_raiser
    try:
        yield
    except Exception as e:
        if is_wrapped_object_deleted_error(e):
            raiser()
        else:
            raise
            

class WindowDeletedError(Exception):
    """Raises when code attempts to manipulate a wx.Window that has been deleted."""
    pass



import bs4
from contextlib import contextmanager
from crystal.browser import MainWindow as RealMainWindow
from crystal.doc.generic import Document, Link
from crystal.doc.html import parse_html_and_links
from crystal.doc.html.basic import BasicDocument
from crystal.doc.html.soup import FAVICON_TYPE_TITLE, HtmlDocument
from crystal.model import (
    Project, Resource, ResourceRevision, ResourceRevisionMetadata, RootResource
)
from crystal.tests.util.asserts import assertEqual, assertRaises
from crystal.tests.util.controls import click_button
from crystal.tests.util.runner import bg_sleep
from crystal.tests.util.server import served_project
from crystal.tests.util.subtests import SubtestsContext, awith_subtests
from crystal.tests.util.wait import DEFAULT_WAIT_PERIOD
from crystal.tests.util.windows import (
    MainWindow, OpenOrCreateDialog, PreferencesDialog,
)
from io import BytesIO
import lxml.html
import os
import tempfile
from textwrap import dedent
from typing import Iterator, List, Tuple
from unittest import skip
from unittest.mock import Mock, patch
from urllib.parse import ParseResult, urlparse, urljoin


# ------------------------------------------------------------------------------
# Tests: HTML Parser Option

async def test_uses_html_parser_specified_in_preferences() -> None:
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        # Define URLs
        home_url = sp.get_request_url('https://xkcd.com/')
        
        async with (await OpenOrCreateDialog.wait_for()).create() as (mw, _):
            project = Project._last_opened_project
            assert project is not None
            
            rr = RootResource(project, 'Home', Resource(project, home_url))
            r = Resource(project, home_url)
            
            # Ensure default HTML parser for new project is lxml
            click_button(mw.preferences_button)
            pd = await PreferencesDialog.wait_for()
            html_parser_title = pd.html_parser_field.Items[pd.html_parser_field.Selection]
            assert 'Fastest - lxml' == html_parser_title
            await pd.ok()
            
            revision_future = r.download(wait_for_embedded=True)
            while not revision_future.done():
                await bg_sleep(DEFAULT_WAIT_PERIOD)
            revision = revision_future.result()
            
            # Ensure expected HTML parser is used
            with _watch_html_parser_usage() as (lxml_parse_func, bs4_parse_func):
                (_, _, _) = revision.document_and_links()
            assert (1, 0) == (lxml_parse_func.call_count, bs4_parse_func.call_count)
            
            # Switch HTML parser
            click_button(mw.preferences_button)
            pd = await PreferencesDialog.wait_for()
            pd.html_parser_field.Selection = pd.html_parser_field.Items.index(
                'Classic - html.parser (bs4)')
            await pd.ok()
            
            # Ensure new HTML parser is used
            with _watch_html_parser_usage() as (lxml_parse_func, bs4_parse_func):
                (_, _, _) = revision.document_and_links()
            assert (0, 1) == (lxml_parse_func.call_count, bs4_parse_func.call_count)


@skip('covered by: test_uses_html_parser_specified_in_preferences')
async def test_defaults_to_lxml_html_parser_for_new_projects() -> None:
    pass


async def test_uses_html_parser_parser_for_classic_projects() -> None:
    # NOTE: The testdata project is a classic project from Crystal <=1.5.0
    with served_project('testdata_xkcd.crystalproj.zip') as sp:
        with RealMainWindow(sp.project) as rmw:
            mw = await MainWindow.wait_for()
            
            click_button(mw.preferences_button)
            pd = await PreferencesDialog.wait_for()
            try:
                html_parser_title = pd.html_parser_field.Items[pd.html_parser_field.Selection]
                assert 'Classic - html.parser (bs4)' == html_parser_title
            finally:
                await pd.ok()


@contextmanager
def _watch_html_parser_usage() -> Iterator[Tuple[Mock, Mock]]:
    with patch('lxml.html.document_fromstring', wraps=lxml.html.document_fromstring) as lxml_parse_func:
        with patch('crystal.util.fastsoup.BeautifulSoup', wraps=bs4.BeautifulSoup) as bs4_parse_func:
            yield (lxml_parse_func, bs4_parse_func)


# ------------------------------------------------------------------------------
# Tests: parse_html_and_links

@skip('not yet automated')
def test_recognizes_link_to_background_image() -> None:
    pass  # type_title = 'Background Image'


@skip('not yet automated')
def test_recognizes_link_to_image_using_src() -> None:
    pass  # type_title = 'Image'


@skip('not yet automated')
def test_recognizes_link_to_iframe() -> None:
    pass  # type_title = 'IFrame'


@skip('not yet automated')
def test_recognizes_link_to_frame_when_no_frameset_present() -> None:
    pass  # type_title = 'Frame'


@skip('not yet automated')
def test_recognizes_link_to_frame_when_frameset_present() -> None:
    pass  # type_title = 'Unknown' (in doc/html/basic.py)


@skip('not yet automated')
def test_recognizes_link_to_form_image() -> None:
    pass  # type_title = 'Form Image'


@skip('not yet automated')
def test_recognizes_unknown_link_type_using_src_attribute() -> None:
    # type_title = 'Unknown Embedded (%s)' % tag_name
    # 
    # ex: <audio>, <embed>, <source>, <track>, <video>
    # via: https://www.w3schools.com/tags/att_src.asp
    pass


@skip('not yet automated')
def test_recognizes_link_to_image_using_srcset() -> None:
    pass  # type_title = 'Image'


@skip('not yet automated')
def test_recognizes_text_link() -> None:
    pass  # type_title = 'Link'


@skip('not yet automated')
def test_recognizes_link_to_stylesheet() -> None:
    # ...using rel stylesheet
    # ...using type css
    # ...using css file extension
    pass  # type_title = 'Stylesheet'


@awith_subtests
async def test_recognizes_explicit_link_to_favicon(subtests: SubtestsContext) -> None:
    with subtests.test('in general'):
        HTML_TEXT = dedent(
            '''
            <!DOCTYPE html>
            <html>
                <head>
                    <title>xkcd: Fast Radio Bursts</title>
                    <link rel="icon" href="/s/919f27.ico" type="image/x-icon"/>
                    <link rel="alternate" type="application/atom+xml" title="Atom 1.0" href="/atom.xml"/>
                    <link rel="alternate" type="application/rss+xml" title="RSS 2.0" href="/rss.xml"/>
                </head>
                <body>
                    ...
                </body>
            </html>
            '''
        ).lstrip('\n')
        assert '919f27.ico' in HTML_TEXT  # ensure has explicit favicon link
        
        with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
            os.rmdir(project_dirpath)
            with Project(project_dirpath) as project:
                rr = ResourceRevision.create_from_response(
                    Resource(project, 'https://www.apple.com/'),
                    metadata=ResourceRevisionMetadata(
                        http_version=10,
                        status_code=200,
                        reason_phrase='OK',
                        headers=[
                            ['Content-Type', 'text/html'],
                        ],
                    ),
                    body_stream=BytesIO(HTML_TEXT.encode('utf-8')))
                (doc, links, _) = rr.document_and_links()
                assert isinstance(doc, HtmlDocument)  # ensure using soup parser
                
                # 1. Ensure detects explicit favicon link
                # 2. Ensure does not insert implicit favicon link
                (favicon_link,) = [
                    link for link in links if
                    link.type_title == FAVICON_TYPE_TITLE and
                    link.relative_url == '/s/919f27.ico'
                ]
                
                # Ensure can rewrite explicit favicon link
                favicon_link.relative_url = '/favicon2.ico'
                assert 'favicon2.ico' in str(doc)
    
    with subtests.test(using='rel shortcut icon'):  # for old IE
        HTML_TEXT = '''<link rel="shortcut icon" href="/s/919f27.ico" type="image/x-icon"/>'''
        (_, links) = _parse_html_and_links(HTML_TEXT)
        (favicon_link,) = [
            link for link in links if
            link.type_title == FAVICON_TYPE_TITLE and
            link.relative_url == '/s/919f27.ico'
        ]
    
    with subtests.test(using='rel icon'):
        HTML_TEXT = '''<link rel="icon" href="/s/919f27.ico" type="image/x-icon"/>'''
        (_, links) = _parse_html_and_links(HTML_TEXT)
        (favicon_link,) = [
            link for link in links if
            link.type_title == FAVICON_TYPE_TITLE and
            link.relative_url == '/s/919f27.ico'
        ]
    
    with subtests.test(using='rel apple-touch-icon'):
        HTML_TEXT = '''<link rel="apple-touch-icon" href="/custom_icon.png">'''
        (_, links) = _parse_html_and_links(HTML_TEXT)
        (favicon_link,) = [
            link for link in links if
            link.type_title == FAVICON_TYPE_TITLE and
            link.relative_url == '/custom_icon.png'
        ]
    
    with subtests.test(using='ico file extension'):
        HTML_TEXT = '''<link href="/myicon.ico" />'''
        (_, links) = _parse_html_and_links(HTML_TEXT)
        (favicon_link,) = [
            link for link in links if
            link.type_title == FAVICON_TYPE_TITLE and
            link.relative_url == '/myicon.ico'
        ]
    
    with subtests.test(using='png file extension'):
        HTML_TEXT = '''<link href="/myicon.png" />'''
        (_, links) = _parse_html_and_links(HTML_TEXT)
        (favicon_link,) = [
            link for link in links if
            link.type_title == FAVICON_TYPE_TITLE and
            link.relative_url == '/myicon.png'
        ]


# NOTE: Logic for recognizing this kind of link currently lives in
#       ResourceRevision.document_and_links() rather than inside
#       any parse_html_and_links() function.
@awith_subtests
async def test_recognizes_implicit_link_to_favicon_from_site_root(subtests: SubtestsContext) -> None:
    HTML_TEXT = dedent(
        '''
        <!DOCTYPE html>
        <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en-US" lang="en-US" prefix="og: http://ogp.me/ns#" class="no-js" data-layout-name="privacy-day">
            <head>
                <meta charset="utf-8"/>
                <link rel="canonical" href="https://www.apple.com/"/>
                <link rel="alternate" href="https://www.apple.com/" hreflang="en-US"/>
                <title>Apple</title>
            </head>
            <body class="page-home ac-nav-overlap globalnav-scrim globalheader-dark nav-dark">
                ...
            </body>
        </html>
        '''
    ).lstrip('\n')
    assert 'favicon.ico' not in HTML_TEXT  # ensure no explicit favicon link
    
    FORCE_BASIC_PARSER_TEXT = '<!-- <frameset></frameset> -->'
    
    with subtests.test(parser='soup'):
        with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
            os.rmdir(project_dirpath)
            with Project(project_dirpath) as project:
                rr = ResourceRevision.create_from_response(
                    Resource(project, 'https://www.apple.com/'),
                    metadata=ResourceRevisionMetadata(
                        http_version=10,
                        status_code=200,
                        reason_phrase='OK',
                        headers=[
                            ['Content-Type', 'text/html'],
                        ],
                    ),
                    body_stream=BytesIO(HTML_TEXT.encode('utf-8')))
                (doc, links, _) = rr.document_and_links()
                assert isinstance(doc, HtmlDocument)  # ensure using soup parser
                
                # Ensure inserts implicit favicon link
                (favicon_link,) = [
                    link for link in links if
                    link.type_title == FAVICON_TYPE_TITLE and
                    link.relative_url == '/favicon.ico'
                ]
                
                # Ensure can rewrite implicit favicon link
                # (in document from soup parser)
                favicon_link.relative_url = '/favicon2.ico'
                assert 'favicon2.ico' in str(doc)
    
    with subtests.test(parser='basic'):
        with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
            os.rmdir(project_dirpath)
            with Project(project_dirpath) as project:
                rr = ResourceRevision.create_from_response(
                    Resource(project, 'https://www.apple.com/'),
                    metadata=ResourceRevisionMetadata(
                        http_version=10,
                        status_code=200,
                        reason_phrase='OK',
                        headers=[
                            ['Content-Type', 'text/html'],
                        ],
                    ),
                    body_stream=BytesIO((HTML_TEXT + FORCE_BASIC_PARSER_TEXT).encode('utf-8')))
                (doc, links, _) = rr.document_and_links()
                assert isinstance(doc, BasicDocument)  # ensure using basic parser
                
                # Ensure inserts implicit favicon link
                (favicon_link,) = [
                    link for link in links if
                    link.type_title == FAVICON_TYPE_TITLE and
                    link.relative_url == '/favicon.ico'
                ]
                
                # Characterize known limitation:
                # Ensure CANNOT rewrite implicit favicon link
                # (in document from basic parser)
                favicon_link.relative_url = '/favicon2.ico'
                assert 'favicon2.ico' not in str(doc)


# TODO: Consider recognizing implicit favicon regardless of whether
#       at site root or not
async def test_does_not_recognize_implicit_link_to_favicon_from_outside_site_root() -> None:
    HTML_TEXT = dedent(
        '''
        <!DOCTYPE html>
        <html class="en-us amr nojs en seg-consumer us" lang="en-US">
            <head>
                <title>Apple Store Online - Apple</title>
                <meta charset="utf-8" />
                <link rel="canonical" href="https://www.apple.com/store" />
                <link rel="alternate" hreflang="en-us" href="https://www.apple.com/store" />
            </head>
            <body class="">
                ...
            </body>
        </html>
        '''
    ).lstrip('\n')
    assert 'favicon.ico' not in HTML_TEXT  # ensure no explicit favicon link
    
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        os.rmdir(project_dirpath)
        with Project(project_dirpath) as project:
            rr = ResourceRevision.create_from_response(
                Resource(project, 'https://www.apple.com/store'),
                metadata=ResourceRevisionMetadata(
                    http_version=10,
                    status_code=200,
                    reason_phrase='OK',
                    headers=[
                        ['Content-Type', 'text/html'],
                    ],
                ),
                body_stream=BytesIO(HTML_TEXT.encode('utf-8')))
            (doc, links, _) = rr.document_and_links()
            
            # Ensure does NOT insert implicit favicon link
            () = [
                link for link in links if
                link.type_title == FAVICON_TYPE_TITLE and
                link.relative_url == '/favicon.ico'
            ]


@skip('not yet automated')
def test_recognizes_link_to_preloaded_resource() -> None:
    pass  # type_title = 'Preload'


@skip('not yet automated')
def test_recognizes_unknown_link_type_using_rel_attribute() -> None:
    # type_title = 'Unknown Link (rel=%s)' % ...
    # 
    # ex: alternate, author, canonical, dns-prefetch, license,
    #     manifest, me, modulepreload, next, pingback, preconnect,
    #     prefetch, prerender, prev, privacy-policy, search,
    #     terms-of-service
    # via: https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes/rel
    pass


@skip('not yet automated')
def test_recognizes_unknown_link_type_using_href_attribute() -> None:
    # type_title = 'Unknown Href (%s)' % tag_name
    # 
    # ex: <none known>
    pass


@skip('not yet automated')
def test_recognizes_onclick_eq_location_eq_ellipsis_link() -> None:
    pass  # type_title = 'Button'


@skip('not yet automated')
def test_recognizes_reference_to_absolute_url_inside_inline_script() -> None:
    pass  # type_title = 'Script Reference'


@skip('not yet automated')
def test_recognizes_reference_to_site_relative_url_inside_inline_script() -> None:
    pass  # type_title = 'Script Reference'


@skip('not yet automated')
def test_recognizes_unknown_attribute_reference_to_absolute_url() -> None:
    pass  # type_title = 'Attribute Reference'


@skip('not yet automated')
def test_does_not_recognize_mailto_or_data_or_javascript_urls_as_links() -> None:
    pass  # see: is_unrewritable_url()


async def test_does_recognize_invalid_relative_urls_as_links() -> None:
    # Ensure test data is detected as an invalid relative URL
    # using native urllib functions
    assert hasattr(urlparse, '_super'), \
        'Expected urlparse() to be patched by patch_urlparse_to_never_raise_exception()'
    assertRaises(
        ValueError,
        lambda: urlparse._super("//*[@id='"))  # type: ignore[attr-defined]
    
    # Ensure test data is detected as an valid relative URL
    # using patched urllib functions
    assertEqual(
        ParseResult(scheme='', netloc='', path="//*[@id='", params='', query='', fragment=''),
        urlparse("//*[@id='"))
    assertEqual(
        "//*[@id='",
        urljoin('https://example.com/', "//*[@id='"))
    
    # Ensure does recognize invalid relative URL as a link in general
    HTML_TEXT = '''<a href="//*[@id='"><a href="#">'''
    (_, links) = _parse_html_and_links(HTML_TEXT)
    (bad_link, good_link) = links
    assert "//*[@id='" == bad_link.relative_url
    assert '#' == good_link.relative_url
    
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        os.rmdir(project_dirpath)
        with Project(project_dirpath) as project:
            rr = ResourceRevision.create_from_response(
                Resource(project, 'https://example.com/'),
                metadata=ResourceRevisionMetadata(
                    http_version=10,
                    status_code=200,
                    reason_phrase='OK',
                    headers=[
                        ['Content-Type', 'text/html'],
                    ],
                ),
                body_stream=BytesIO(HTML_TEXT.encode('utf-8')))
            (doc, links, _) = rr.document_and_links()
            
            # Ensure does recognize invalid relative URL as a document link
            (bad_link, good_link, favicon_link) = links
            assert "//*[@id='" == bad_link.relative_url
            assert '#' == good_link.relative_url
            assert '/favicon.ico' == favicon_link.relative_url


async def test_recognizes_http_redirect_as_a_link() -> None:
    with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
        os.rmdir(project_dirpath)
        with Project(project_dirpath) as project:
            rr = ResourceRevision.create_from_response(
                Resource(project, 'https://www.xkcd.com/'),
                metadata=ResourceRevisionMetadata(
                    http_version=10,
                    status_code=301,
                    reason_phrase='Moved Permanently',
                    headers=[
                        ['Location', 'https://xkcd.com/'],
                    ],
                ),
                body_stream=BytesIO(b''))
            (doc, links, _) = rr.document_and_links()
            
            # Ensure inserts redirect link
            (redirect_link,) = [
                link for link in links if
                link.type_title == 'Redirect' and
                link.relative_url == 'https://xkcd.com/' and
                link.embedded
            ]


@skip('fails: not yet implemented')
def test_recognizes_meta_http_equiv_refresh_as_a_link() -> None:
    # see: https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meta#examples
    pass  # type_title = 'Redirect' (proposed)


@skip('not yet automated')
def test_recognizes_links_defined_by_plugins() -> None:
    # ex: plugins_minbaker.postprocess_document_and_links()
    pass


def _parse_html_and_links(html: str) -> Tuple[Document, List[Link]]:
    doc_and_links = parse_html_and_links(html.encode('utf-8'), 'utf-8', 'lxml')
    assert doc_and_links is not None
    return doc_and_links


# ------------------------------------------------------------------------------



from __future__ import annotations

from contextlib import asynccontextmanager
from crystal.browser.new_root_url import NewRootUrlDialog as RealNewRootUrlDialog
from crystal.model import Project
from crystal.tests.util.controls import (
    click_button, file_dialog_returning,
    select_menuitem_now, TreeItem
)
from crystal.tests.util.runner import bg_sleep, pump_wx_events
from crystal.tests.util.tasks import first_task_title_progression
from crystal.tests.util.wait import (
    tree_has_no_children_condition,
    wait_for, WaitTimedOut, window_condition, not_condition, or_condition
)
from crystal.util.xos import is_mac_os
import os.path
import re
import sys
import tempfile
import traceback
from typing import AsyncIterator, Callable, Optional, Tuple, TYPE_CHECKING
import wx

if TYPE_CHECKING:
    import wx.adv


# ------------------------------------------------------------------------------
# Window & Dialog Abstractions

class OpenOrCreateDialog:
    # NOTE: 10.0 isn't long enough for Windows test runners on GitHub Actions
    _TIMEOUT_FOR_OPEN_MAIN_WINDOW = 12.0
    
    open_or_create_project_dialog: wx.Dialog
    open_as_readonly: wx.CheckBox
    open_button: wx.Button
    create_button: wx.Button
    
    @staticmethod
    async def wait_for(timeout: Optional[float]=None) -> OpenOrCreateDialog:
        self = OpenOrCreateDialog(ready=True)
        open_or_create_project_dialog = await wait_for(
            window_condition('cr-open-or-create-project'),
            timeout=timeout,
            stacklevel_extra=1,
        )  # type: wx.Window
        assert isinstance(open_or_create_project_dialog, wx.Dialog)
        self.open_or_create_project_dialog = open_or_create_project_dialog
        self.open_as_readonly = self.open_or_create_project_dialog.FindWindow(name=
            'cr-open-or-create-project__checkbox')
        assert isinstance(self.open_as_readonly, wx.CheckBox)
        self.open_button = self.open_or_create_project_dialog.FindWindow(id=wx.ID_NO)
        assert isinstance(self.open_button, wx.Button)
        self.create_button = self.open_or_create_project_dialog.FindWindow(id=wx.ID_YES)
        assert isinstance(self.create_button, wx.Button)
        return self
    
    def __init__(self, *, ready: bool=False) -> None:
        assert ready, 'Did you mean to use OpenOrCreateDialog.wait_for()?'
    
    @asynccontextmanager
    async def create(self, 
            project_dirpath: Optional[str]=None,
            *, autoclose: bool=True,
            ) -> AsyncIterator[Tuple[MainWindow, str]]:
        if project_dirpath is None:
            with tempfile.TemporaryDirectory(suffix='.crystalproj') as project_dirpath:
                assert project_dirpath is not None
                async with self.create(project_dirpath, autoclose=autoclose) as (mw, project_dirpath):
                    yield (mw, project_dirpath)
            return
        
        mw = await self.create_and_leave_open(project_dirpath)
        
        exc_info_while_close = None
        try:
            yield (mw, project_dirpath)
        except:
            exc_info_while_close = sys.exc_info()
            raise
        finally:
            if autoclose:
                await mw.close(exc_info_while_close)
    
    async def create_and_leave_open(self, project_dirpath: str) -> MainWindow:
        old_opened_project = Project._last_opened_project  # capture
        try:
            with file_dialog_returning(project_dirpath):
                click_button(self.create_button)
                
                mw = await MainWindow.wait_for(timeout=self._TIMEOUT_FOR_OPEN_MAIN_WINDOW)
                return mw
        except:
            # Close any project that was opened
            new_opened_project = Project._last_opened_project  # capture
            if new_opened_project != old_opened_project:
                if new_opened_project is not None:
                    new_opened_project.close()
            
            raise
    
    @asynccontextmanager
    async def open(self, 
            project_dirpath: str, 
            *, readonly: Optional[bool]=None,
            autoclose: bool=True,
            using_crystalopen: bool=False,
            ) -> AsyncIterator[MainWindow]:
        if readonly is not None:
            self.open_as_readonly.Value = readonly
        
        if using_crystalopen:
            itempath_to_open = os.path.join(project_dirpath, Project._OPENER_DEFAULT_FILENAME)
        else:
            itempath_to_open = project_dirpath
        
        with file_dialog_returning(itempath_to_open):
            click_button(self.open_button)
            
            mw = await MainWindow.wait_for(timeout=self._TIMEOUT_FOR_OPEN_MAIN_WINDOW)
        
        exc_info_while_close = None
        try:
            yield mw
        except:
            exc_info_while_close = sys.exc_info()
            raise
        finally:
            if autoclose:
                await mw.close(exc_info_while_close)
    
    async def start_opening(self, project_dirpath: str, *, next_window_name: str) -> None:
        with file_dialog_returning(project_dirpath):
            click_button(self.open_button)
            
            await wait_for(
                or_condition(
                    window_condition('cr-opening-project'),
                    window_condition('cr-main-window'),
                    window_condition(next_window_name)),
                stacklevel_extra=1)


class MainWindow:
    _connect_timeout: Optional[float]
    
    main_window: wx.Frame
    entity_tree: 'EntityTree'
    new_root_url_button: wx.Button
    new_group_button: wx.Button
    edit_button: wx.Button
    forget_button: wx.Button
    download_button: wx.Button
    update_members_button: wx.Button
    view_button: wx.Button
    task_tree: wx.TreeCtrl
    preferences_button: wx.Button
    read_write_icon: wx.StaticText
    
    @staticmethod
    async def wait_for(*, timeout: Optional[float]=None) -> MainWindow:
        self = MainWindow(ready=True)
        self._connect_timeout = timeout
        await self._connect()
        return self
    
    async def _connect(self) -> None:
        self.main_window = await wait_for(
            window_condition('cr-main-window'),
            timeout=self._connect_timeout,
            stacklevel_extra=1)
        assert isinstance(self.main_window, wx.Frame)
        entity_tree_window = self.main_window.FindWindow(name='cr-entity-tree')
        assert isinstance(entity_tree_window, wx.TreeCtrl)
        self.entity_tree = EntityTree(entity_tree_window)
        self.new_root_url_button = self.main_window.FindWindow(name='cr-add-url-button')
        assert isinstance(self.new_root_url_button, wx.Button)
        self.new_group_button = self.main_window.FindWindow(name='cr-add-group-button')
        assert isinstance(self.new_group_button, wx.Button)
        self.edit_button = self.main_window.FindWindow(name='cr-edit-button')
        assert isinstance(self.edit_button, wx.Button)
        self.forget_button = self.main_window.FindWindow(name='cr-forget-button')
        assert isinstance(self.forget_button, wx.Button)
        self.download_button = self.main_window.FindWindow(name='cr-download-button')
        assert isinstance(self.download_button, wx.Button)
        self.update_members_button = self.main_window.FindWindow(name='cr-update-members-button')
        assert isinstance(self.update_members_button, wx.Button)
        self.view_button = self.main_window.FindWindow(name='cr-view-button')
        assert isinstance(self.view_button, wx.Button)
        self.task_tree = self.main_window.FindWindow(name='cr-task-tree')
        assert isinstance(self.task_tree, wx.TreeCtrl)
        self.preferences_button = self.main_window.FindWindow(name='cr-preferences-button')
        assert isinstance(self.preferences_button, wx.Button)
        self.read_write_icon = self.main_window.FindWindow(name='cr-read-write-icon')
        assert isinstance(self.read_write_icon, wx.StaticText)
    
    def __init__(self, *, ready: bool=False) -> None:
        assert ready, 'Did you mean to use MainWindow.wait_for()?'
    
    # === Menubar ===
    
    @property
    def entity_menu(self) -> wx.Menu:
        mb = self.main_window.MenuBar
        entity_menu_index = mb.FindMenu('Entity')
        assert entity_menu_index != wx.NOT_FOUND
        entity_menu = mb.GetMenu(entity_menu_index)
        return entity_menu
    
    # === Properties ===
    
    @property
    def readonly(self) -> bool:
        label = self.read_write_icon.Label  # cache
        if label in ['🔒', 'Read only']:
            return True
        elif label in ['✏️', 'Writable']:
            return False
        else:
            raise AssertionError()
    
    # === Operations ===
    
    async def click_download_button(self, *, immediate_finish_ok: bool=False) -> None:
        """
        Clicks the "Download" button and waits for it to finish starting
        a download task.
        """
        task_root_ti = TreeItem.GetRootItem(self.task_tree)
        assert task_root_ti is not None
        
        old_task_count = len(task_root_ti.Children)
        
        click_button(self.download_button)
        def task_count_changed_condition() -> Optional[int]:
            assert task_root_ti is not None
            new_task_count = len(task_root_ti.Children)
            if new_task_count != old_task_count:
                return new_task_count
            else:
                return None
        try:
            await wait_for(task_count_changed_condition, stacklevel_extra=1)
        except WaitTimedOut:
            if immediate_finish_ok:
                # TOOD: We ended up waiting until the entire timeout expired.
                #       Find a more efficient way to detect that the download
                #       started and immediately finished.
                return
            else:
                raise
    
    # === Close ===
    
    CLOSE_TIMEOUT = 4.0
    
    async def close(self, exc_info=None) -> None:
        # Try wait for any lingering tasks to complete.
        # 
        # Does workaround: https://github.com/davidfstr/Crystal-Web-Archiver/issues/74
        try:
            await wait_for(
                tree_has_no_children_condition(self.task_tree),
                timeout=self.CLOSE_TIMEOUT,  # wait only briefly
                stacklevel_extra=1)
        except WaitTimedOut:
            first_task_title = first_task_title_progression(self.task_tree)()
            print(f'*** MainWindow: Closing while tasks are still running. May deadlock! Current task: {first_task_title}')
            
            # Print traceback or current exception that is being handled immediately
            # because impending deadlock may prevent traceback from being printed
            # in the usual manner
            if exc_info is None:
                traceback.print_stack()
            else:
                traceback.print_exception(*exc_info)
            
            # (continue)
        
        self.main_window.Close()
        await self.wait_for_close()
    
    async def close_with_menuitem(self) -> None:
        close_menuitem = self.main_window.MenuBar.FindItemById(wx.ID_CLOSE)
        wx.PostEvent(close_menuitem.Menu, wx.CommandEvent(wx.EVT_MENU.typeId, close_menuitem.Id))
        await self.wait_for_close()
    
    async def wait_for_close(self) -> None:
        await wait_for(lambda: self.main_window.IsBeingDeleted)
        await wait_for(lambda: not self.main_window.IsShown)
        await wait_for(
            not_condition(window_condition('cr-main-window')),
            timeout=4.0,  # 2.0s isn't long enough for macOS test runners on GitHub Actions
            stacklevel_extra=1,
        )
        await OpenOrCreateDialog.wait_for(
            timeout=4.0  # 2.0s isn't long enough for macOS test runners on GitHub Actions
        )
    
    @asynccontextmanager
    async def temporarily_closed(self, project_dirpath: str) -> AsyncIterator[None]:
        await self.close()
        try:
            yield
        finally:
            async with (await OpenOrCreateDialog.wait_for()).open(project_dirpath, autoclose=False):
                pass
            await self._connect()  # reconnect self
    
    async def open_preferences_with_menuitem(self) -> PreferencesDialog:
        prefs_menuitem = self.main_window.MenuBar.FindItemById(wx.ID_PREFERENCES)
        wx.PostEvent(self.main_window, wx.CommandEvent(wx.EVT_MENU.typeId, prefs_menuitem.Id))
        return await PreferencesDialog.wait_for()
    
    async def quit_with_menuitem(self) -> None:
        quit_menuitem = self.main_window.MenuBar.FindItemById(wx.ID_EXIT)
        wx.PostEvent(self.main_window, wx.CommandEvent(wx.EVT_MENU.typeId, quit_menuitem.Id))
        
        await self.wait_for_close()


class NewRootUrlDialog:
    _dialog: wx.Dialog
    _controller: RealNewRootUrlDialog
    url_field: wx.TextCtrl
    url_cleaner_spinner: wx.ActivityIndicator
    name_field: wx.TextCtrl
    ok_button: wx.Button
    cancel_button: wx.Button
    
    options_button: wx.Button
    set_as_default_domain_checkbox: wx.CheckBox
    set_as_default_directory_checkbox: wx.CheckBox
    
    @staticmethod
    async def wait_for() -> NewRootUrlDialog:
        self = NewRootUrlDialog(ready=True)
        self._dialog = await wait_for(window_condition('cr-new-root-url-dialog'), stacklevel_extra=1)
        assert isinstance(self._dialog, wx.Dialog)
        assert RealNewRootUrlDialog._last_opened is not None
        self._controller = RealNewRootUrlDialog._last_opened
        self.url_field = self._dialog.FindWindow(name='cr-new-root-url-dialog__url-field')
        assert isinstance(self.url_field, wx.TextCtrl)
        self.url_cleaner_spinner = self._dialog.FindWindow(name='cr-new-root-url-dialog__url-cleaner-spinner')
        assert isinstance(self.url_cleaner_spinner, wx.ActivityIndicator)
        self.name_field = self._dialog.FindWindow(name='cr-new-root-url-dialog__name-field')
        assert isinstance(self.name_field, wx.TextCtrl)
        self.ok_button = self._dialog.FindWindow(id=wx.ID_NEW)
        if self.ok_button is None:
            self.ok_button = self._dialog.FindWindow(id=wx.ID_SAVE)
        assert isinstance(self.ok_button, wx.Button)
        self.cancel_button = self._dialog.FindWindow(id=wx.ID_CANCEL)
        assert isinstance(self.cancel_button, wx.Button)
        self.options_button = self._dialog.FindWindow(id=wx.ID_MORE)
        assert isinstance(self.options_button, wx.Button)
        self.set_as_default_domain_checkbox = self._dialog.FindWindow(name='cr-new-root-url-dialog__set-as-default-domain-checkbox')
        assert isinstance(self.set_as_default_domain_checkbox, wx.CheckBox)
        self.set_as_default_directory_checkbox = self._dialog.FindWindow(name='cr-new-root-url-dialog__set-as-default-directory-checkbox')
        assert isinstance(self.set_as_default_directory_checkbox, wx.CheckBox)
        return self
    
    def __init__(self, *, ready: bool=False) -> None:
        assert ready, 'Did you mean to use NewRootUrlDialog.wait_for()?'
    
    @property
    def shown(self) -> bool:
        if self._controller._is_destroying_or_destroyed:
            return False
        return self._dialog.IsShown()
    
    @property
    def url_field_focused(self) -> bool:
        return self._controller._url_field_focused
    
    async def ok(self) -> None:
        click_button(self.ok_button)
        await wait_for(not_condition(window_condition('cr-new-root-url-dialog')), stacklevel_extra=1)
    
    async def cancel(self) -> None:
        click_button(self.cancel_button)
        await wait_for(not_condition(window_condition('cr-new-root-url-dialog')), stacklevel_extra=1)
    
    def do_not_set_default_url_prefix(self) -> None:
        """
        Configures the URL being created to NOT also set it as the default domain.
        
        Several test infrastructure methods (like TreeItem.find_child) are easier
        to use when there is no Default URL Prefix set, so some tests intentionally
        avoid setting a Default URL Prefix.
        """
        click_button(self.options_button)
        if self.set_as_default_domain_checkbox.Value:
            self.set_as_default_domain_checkbox.Value = False
        
        assert False == self.set_as_default_domain_checkbox.Value
        assert False == self.set_as_default_directory_checkbox.Value


class NewGroupDialog:
    _NONE_SOURCE_TITLE = 'none'
    _SOURCE_TITLE_RE = re.compile(r'^(?:[^a-zA-Z0-9]+ )?(.*?)(?: - (.*?))? *$')
    
    name_field: wx.TextCtrl
    pattern_field: wx.TextCtrl
    source_field: wx.Choice
    preview_members_pane: Optional[wx.CollapsiblePane]
    preview_members_list: wx.ListBox
    cancel_button: wx.Button
    ok_button: wx.Button
    
    @staticmethod
    async def wait_for() -> NewGroupDialog:
        self = NewGroupDialog(ready=True)
        add_group_dialog = await wait_for(
            NewGroupDialog.window_condition(),
            timeout=5.0,  # 4.2s observed for macOS test runners on GitHub Actions
            stacklevel_extra=1,
        )  # type: wx.Window
        self.name_field = add_group_dialog.FindWindow(name='cr-new-group-dialog__name-field')
        assert isinstance(self.name_field, wx.TextCtrl)
        self.pattern_field = add_group_dialog.FindWindow(name='cr-new-group-dialog__pattern-field')
        assert isinstance(self.pattern_field, wx.TextCtrl)
        self.source_field = add_group_dialog.FindWindow(name='cr-new-group-dialog__source-field')
        assert isinstance(self.source_field, wx.Choice)
        self.preview_members_pane = add_group_dialog.FindWindow(name='cr-new-group-dialog__preview-members')
        assert (
            self.preview_members_pane is None or
            isinstance(self.preview_members_pane, wx.CollapsiblePane)
        )
        self.preview_members_list = add_group_dialog.FindWindow(name='cr-new-group-dialog__preview-members__list')
        assert isinstance(self.preview_members_list, wx.ListBox)
        self.cancel_button = add_group_dialog.FindWindow(id=wx.ID_CANCEL)
        assert isinstance(self.cancel_button, wx.Button)
        self.ok_button = add_group_dialog.FindWindow(id=wx.ID_NEW)
        if self.ok_button is None:
            self.ok_button = add_group_dialog.FindWindow(id=wx.ID_SAVE)
        assert isinstance(self.ok_button, wx.Button)
        return self
    
    @staticmethod
    def window_condition() -> Callable[[], Optional[wx.Window]]:
        return window_condition('cr-new-group-dialog')
    
    def __init__(self, *, ready: bool=False) -> None:
        assert ready, 'Did you mean to use NewGroupDialog.wait_for()?'
    
    # TODO: Rename -> source_name
    def _get_source(self) -> Optional[str]:
        selection_ci = self.source_field.GetSelection()
        if selection_ci == wx.NOT_FOUND:
            return None
        source_title = self.source_field.GetString(selection_ci)
        if source_title == self._NONE_SOURCE_TITLE:
            return None
        m = self._SOURCE_TITLE_RE.fullmatch(source_title)
        assert m is not None
        source_name = m.group(2)
        if source_name is not None:
            return source_name
        else:
            # If the referenced source has no name, allow referring to it by its display URL
            cur_source_display_url = m.group(1)
            assert cur_source_display_url is not None
            return cur_source_display_url
            
    def _set_source(self, source_name: Optional[str]) -> None:
        if source_name is None:
            selection_ci = 0
        else:
            for (selection_ci, source_title) in enumerate(self.source_field.GetStrings()):
                if selection_ci == 0:
                    continue
                m = self._SOURCE_TITLE_RE.fullmatch(source_title)
                assert m is not None
                cur_source_name = m.group(2)
                if cur_source_name is not None:
                    if cur_source_name == source_name:
                        break
                else:
                    # If the referenced source has no name, allow referring to it by its display URL
                    cur_source_display_url = m.group(1)
                    assert cur_source_display_url is not None
                    if cur_source_display_url == source_name:
                        break
            else:
                raise ValueError(f'Source not found: {source_name}')
        self.source_field.SetSelection(selection_ci)
    source = property(_get_source, _set_source)
    
    async def ok(self) -> None:
        click_button(self.ok_button)
        await wait_for(not_condition(window_condition('cr-new-group-dialog')), stacklevel_extra=1)
    
    async def cancel(self) -> None:
        click_button(self.cancel_button)
        await wait_for(not_condition(window_condition('cr-new-group-dialog')), stacklevel_extra=1)


class PreferencesDialog:
    html_parser_field: wx.Choice
    stale_before_checkbox: wx.CheckBox
    stale_before_date_picker: 'wx.adv.DatePickerCtrl'
    cookie_field: wx.ComboBox
    ok_button: wx.Button
    
    @staticmethod
    async def wait_for() -> PreferencesDialog:
        import wx.adv
        
        self = PreferencesDialog(ready=True)
        preferences_dialog = await wait_for(
            window_condition('cr-preferences-dialog'),
            timeout=4.0,  # 2.0s isn't long enough for macOS test runners on GitHub Actions
            stacklevel_extra=1,
        )  # type: wx.Window
        self.html_parser_field = preferences_dialog.FindWindow(name=
            'cr-preferences-dialog__html-parser-field')
        assert isinstance(self.html_parser_field, wx.Choice)
        self.stale_before_checkbox = preferences_dialog.FindWindow(name=
            'cr-preferences-dialog__stale-before-checkbox')
        assert isinstance(self.stale_before_checkbox, wx.CheckBox)
        self.stale_before_date_picker = preferences_dialog.FindWindow(name=
            'cr-preferences-dialog__stale-before-date-picker')
        assert isinstance(self.stale_before_date_picker, wx.adv.DatePickerCtrl)
        self.cookie_field = preferences_dialog.FindWindow(name=
            'cr-preferences-dialog__cookie-field')
        assert isinstance(self.cookie_field, wx.ComboBox)
        self.ok_button = preferences_dialog.FindWindow(id=wx.ID_OK)
        assert isinstance(self.ok_button, wx.Button)
        return self
    
    def __init__(self, *, ready: bool=False) -> None:
        assert ready, 'Did you mean to use PreferencesDialog.wait_for()?'
    
    async def ok(self) -> None:
        click_button(self.ok_button)
        await wait_for(not_condition(window_condition('cr-preferences-dialog')), stacklevel_extra=1)


# ------------------------------------------------------------------------------
# Panel Abstractions

class EntityTree:
    def __init__(self, window: wx.TreeCtrl) -> None:
        self.window = window
    
    async def get_tree_item_icon_tooltip(self, tree_item: TreeItem) -> Optional[str]:
        if tree_item.tree != self.window:
            raise ValueError()
        return tree_item.Tooltip('icon')
    
    @staticmethod
    async def assert_tree_item_icon_tooltip_contains(ti: TreeItem, value: str) -> None:
        tooltip = await (EntityTree(ti.tree).get_tree_item_icon_tooltip(ti))
        assert tooltip is not None and value in tooltip, \
            f'Expected tooltip to contain {value!r}, but it was: {tooltip!r}'
    
    async def set_default_domain_to_entity_at_tree_item(self, tree_item: TreeItem) -> None:
        """
        Raises:
        * MenuitemMissingError
        * MenuitemDisabledError
        """
        await self._choose_action_for_entity_at_tree_item(tree_item, 'Set As Default Domain')
    
    async def set_default_directory_to_entity_at_tree_item(self, tree_item: TreeItem) -> None:
        """
        Raises:
        * MenuitemMissingError
        * MenuitemDisabledError
        """
        await self._choose_action_for_entity_at_tree_item(tree_item, 'Set As Default Directory')
    
    async def clear_default_domain_from_entity_at_tree_item(self, tree_item: TreeItem) -> None:
        """
        Raises:
        * MenuitemMissingError
        * MenuitemDisabledError
        """
        await self._choose_action_for_entity_at_tree_item(tree_item, 'Clear Default Domain')
    
    async def clear_default_directory_from_entity_at_tree_item(self, tree_item: TreeItem) -> None:
        """
        Raises:
        * MenuitemMissingError
        * MenuitemDisabledError
        """
        await self._choose_action_for_entity_at_tree_item(tree_item, 'Clear Default Directory')
    
    async def _choose_action_for_entity_at_tree_item(self,
            tree_item: TreeItem,
            action_prefix: str) -> None:
        """
        Raises:
        * MenuitemMissingError
        * MenuitemDisabledError
        """
        def show_popup(menu: wx.Menu) -> None:
            try:
                (set_prefix_menuitem,) = [
                    mi for mi in menu.MenuItems
                    if mi.ItemLabelText.startswith(action_prefix)
                ]
            except ValueError:  # not enough values to unpack
                raise MenuitemMissingError()
            if not set_prefix_menuitem.Enabled:
                raise MenuitemDisabledError()
            select_menuitem_now(menu, set_prefix_menuitem.Id)
        await tree_item.right_click_showing_popup_menu(show_popup)


class MenuitemMissingError(ValueError):
    pass


class MenuitemDisabledError(ValueError):
    pass


# ------------------------------------------------------------------------------



"""
Parses XML documents.
"""

from bs4 import BeautifulSoup
from crystal.doc.generic import Document, Link
from crystal.doc.html.soup import HtmlDocument, HtmlLink
from crystal.util.fastsoup import BeautifulFastSoup
import re
from typing import BinaryIO, List, Optional, Tuple


def parse_xml_and_links(
        xml_bytes: BinaryIO,
        declared_charset: Optional[str]=None
        ) -> Optional[Tuple[Document, List[Link]]]:
    """
    Parses an XML document, returning a FastSoup object that can be
    examined through a BeautifulSoup-compatible API.
    
    Returns None if there was a parsing error.
    """
    try:
        xml = BeautifulFastSoup(BeautifulSoup(
            xml_bytes,
            from_encoding=declared_charset,
            features='xml',
        ))
    except Exception as e:
        return None
    
    links = []
    
    # <link>*</link> (from RSS feed)
    # <link href=*> (from Atom feed)
    type_title = 'Link'
    title = None
    embedded = False
    for tag in xml.find_all('link'):
        if xml.tag_string(tag) not in [None, '']:
            links.append(HtmlLink.create_from_tag(tag, xml, 'string', type_title, title, embedded))
        if 'href' in  xml.tag_attrs(tag):  # usually also has: rel="alternate"
            links.append(HtmlLink.create_from_tag(tag, xml, 'href', type_title, title, embedded))
    
    links_ = links  # type: List[Link]  # type: ignore[assignment]  # allow List[HtmlLink] to be converted
    return (HtmlDocument(xml, is_html=False), links_)




import atexit
from contextlib import AbstractContextManager, contextmanager, nullcontext
import cProfile
from functools import wraps
import inspect
import sys
import threading
import time
from typing import Callable, Iterator, Optional, Union


_excluded_delta_time_stack = threading.local()

@contextmanager
def warn_if_slow(
        title: str,
        max_duration: float,
        message: Union[Callable[[], str], str],
        *, enabled: bool=True
        ) -> Iterator[None]:
    """
    Context that profiles the runtime of its enclosed code and
    upon exit prints a warning if the runtime exceeds the specified `max_duration`.
    
    If a warn_if_slow() context is nested within another warn_if_slow() context,
    the runtime of the inner context will be excluded from the runtime of the
    outer context.
    
    If a ignore_runtime_from_enclosing_warn_if_slow() context is nested within
    a warn_if_slow() context, the runtime of the inner context will be excluded
    from the runtime of the outer context.
    """
    if not enabled:
        yield
        return
    if not (max_duration > 0):
        raise ValueError()
    
    if not hasattr(_excluded_delta_time_stack, 'value'):
        _excluded_delta_time_stack.value = []
    _excluded_delta_time_stack.value.append(0)
    
    start_time = time.time()  # capture
    try:
        yield
    finally:
        end_time = time.time()  # capture
        
        excluded_delta_time = _excluded_delta_time_stack.value.pop()
        
        delta_time = end_time - start_time  # cache
        if (delta_time - excluded_delta_time) > max_duration:
            message_str = message() if callable(message) else message
            assert isinstance(message_str, str)
            excluded_part = (
                ' (%.02fs excluded)' % (excluded_delta_time,)
                if excluded_delta_time > 0
                else ''
            )
            print("*** %s took %.02fs%s to execute: %s" % (
                title,
                delta_time,
                excluded_part,
                message_str,
            ), file=sys.stderr)
            
            # Exclude delta_time from any enclosing calls of warn_if_slow()
            for (i, _) in enumerate(_excluded_delta_time_stack.value):
                _excluded_delta_time_stack.value[i] += delta_time


@contextmanager
def ignore_runtime_from_enclosing_warn_if_slow() -> Iterator[None]:
    """
    Context that excluded its runtime from that of any enclosing
    warn_if_slow() context.
    """
    if not hasattr(_excluded_delta_time_stack, 'value'):
        _excluded_delta_time_stack.value = []
    
    start_time = time.time()  # capture
    try:
        yield
    finally:
        end_time = time.time()  # capture
        delta_time = end_time - start_time  # cache
        
        # Exclude delta_time from any enclosing calls of warn_if_slow()
        for (i, _) in enumerate(_excluded_delta_time_stack.value):
            _excluded_delta_time_stack.value[i] += delta_time


def create_profiled_callable(title: str, max_duration: float, callable: Callable, *args) -> Callable:
    """
    Decorates the specified callable such that it prints
    a warning to the console if its runtime is long.
    """
    @wraps(callable)
    def profiled_callable() -> None:
        def message_func() -> str:
            # TODO: Remove support for nested callable objects that are
            #       tracked using a "callable" attribute. I believe this
            #       is dead code.
            root_callable = callable
            while hasattr(root_callable, 'callable'):
                root_callable = root_callable.callable  # type: ignore[attr-defined]
            
            try:
                file = inspect.getsourcefile(root_callable)
            except Exception:
                file = '?'
            try:
                start_line_number = str(inspect.getsourcelines(root_callable)[-1])
            except Exception:
                start_line_number = '?'
            return '%s @ [%s:%s]' % (
                root_callable,
                file,
                start_line_number
            )
            
        with warn_if_slow(title, max_duration, message_func):
            return callable(*args)
    return profiled_callable


def create_profiling_context(
        stats_filepath: str,
        *, enabled: bool=True,
        ) -> 'AbstractContextManager[Optional[cProfile.Profile]]':
    """
    Creates a cProfile profiling context.
    Within the context all function calls are timed.
    The content can be entered and exited mutliple times.
    
    Just before the program exits, a .stats file is written to the specified
    filepath. This file can be analyzed/visualized with the PyPI "flameprof"
    module and the standard library "pstats" module.
    """
    if enabled:
        profiling_context = cProfile.Profile()  # type: AbstractContextManager[Optional[cProfile.Profile]]
        @atexit.register
        def dump_stats() -> None:  # type: ignore[misc]
            profiler = profiling_context
            assert isinstance(profiler, cProfile.Profile)
            profiler.dump_stats(stats_filepath)
    else:
        profiling_context = nullcontext(enter_result=None)
    return profiling_context



try:
    # Python 3.9+
    from functools import cache  # type: ignore[attr-defined]
except ImportError:
    # Python 3.8
    from functools import lru_cache
    cache = lambda func: lru_cache(maxsize=None)(func)



"""
Parses CSS documents.
"""

from crystal.doc.generic import Document, Link
import tinycss2
from tinycss2 import ast
from tinycss2.serializer import serialize_string_value, serialize_url
from typing import List, Optional, Tuple


def parse_css_and_links(
        body_bytes: bytes, 
        declared_charset: Optional[str]=None
        ) -> 'Tuple[CssDocument, List[Link]]':
    (rules, encoding) = tinycss2.parse_stylesheet_bytes(
        body_bytes,
        protocol_encoding=declared_charset)
    
    links = []  # type: List[Link]
    for rule in rules:
        if isinstance(rule, ast.QualifiedRule) or isinstance(rule, ast.AtRule):
            if rule.content is not None:  # has been observed as None in the wild sometimes
                for token in rule.content:
                    # url(**)
                    if isinstance(token, ast.URLToken):
                        links.append(UrlTokenLink(token))
                    
                    # url("**")
                    elif isinstance(token, ast.FunctionBlock):
                        if (token.lower_name == 'url' and 
                                len(token.arguments) == 1 and 
                                isinstance(token.arguments[0], ast.StringToken)):
                            links.append(UrlFunctionLink(token.arguments[0]))
        
        # @import "**";
        if isinstance(rule, ast.AtRule) and rule.at_keyword == 'import':
            for token in rule.prelude:
                if isinstance(token, ast.StringToken):
                    links.append(StringTokenLink(token))
    
    return (CssDocument(rules), links)


class CssDocument(Document):
    def __init__(self, rules) -> None:
        self._rules = rules
    
    def __str__(self) -> str:
        return tinycss2.serialize(self._rules)


class UrlTokenLink(Link):
    def __init__(self, token: ast.URLToken) -> None:
        self._token = token
        
        self.title = None
        self.type_title = 'CSS URL Reference'
        self.embedded = True
    
    def _get_relative_url(self) -> str:
        return self._token.value
    def _set_relative_url(self, url: str) -> None:
        self._token.value = url
        self._token.representation = 'url(%s)' % serialize_url(url)
    relative_url = property(_get_relative_url, _set_relative_url)


class UrlFunctionLink(Link):
    def __init__(self, token: ast.StringToken) -> None:
        self._token = token
        
        self.title = None
        self.type_title = 'CSS URL Reference'
        self.embedded = True
    
    def _get_relative_url(self) -> str:
        return self._token.value
    def _set_relative_url(self, url: str) -> None:
        self._token.value = url
        self._token.representation = serialize_string_value(url)
    relative_url = property(_get_relative_url, _set_relative_url)


class StringTokenLink(Link):
    def __init__(self, token: ast.StringToken) -> None:
        self._token = token
        
        self.title = None
        self.type_title = 'CSS @import'
        self.embedded = True
    
    def _get_relative_url(self) -> str:
        return self._token.value
    def _set_relative_url(self, url: str) -> None:
        self._token.value = url
        self._token.representation = '"%s"' % serialize_string_value(url)
    relative_url = property(_get_relative_url, _set_relative_url)



from crystal.tests.util.skip import skipTest
from crystal.util.xos import is_linux, is_windows
from functools import wraps
import platform
from unittest import TestCase


# ------------------------------------------------------------------------------
# Utility: Skip on Windows

def skip_on_windows(func=lambda: None):
    """Decorator for tests that should be skipped on Windows."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        if is_windows():
            skipTest('not supported on Windows')
        func(*args, **kwargs)
    return wrapper


# ------------------------------------------------------------------------------
